/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: appos.c
*    Creation Date: 17 September 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.2 $
*        $Date: 2005/08/15 15:00:39 $
*
****************************************************************************
*    Description:
*
*     Broadcom Application-specific Operating System Functions. This file should
*     only contain functions which are necessary to supplement the bcmos
*     functionality to implement application specific purposes.
*
****************************************************************************/

#include <hausenv.h>
#include <string.h>
///#include <xbcmos.h>
#include <appos.h>
#include <apposCfg.h>


/*
*****************************************************************************
** TYPE DEFINITIONS and CONSTANTS
*****************************************************************************
*/

/* The reset granularity is the rate at which blocking OS calls check if
 * an application reset has been initiated. If it has, the OS call will return
 * an AOS_STATUS_RESET error code. */
#define RESET_GRANULARITY_TICKS  APPOS_CFG_RESET_GRANULARITY

#define AOS_TASKRESET_SEM        "aoTR"
#define AOS_APPRESET_SEM         "aoAR"
#define AOS_TCB_MUTEX            "aoTM"
#define AOS_ROOTTASK_NAME        "aoRT"
#define AOS_APPTASK_NAME         "aoAP"

/* Maximum number of tasks in the system */
#define MAX_TCB                  APPOS_CFG_MAX_TCB


#define PUBLIC
#define PRIVATE static


/* APPOS task states */
typedef enum
{
   AOS_TASKSTATE_NORMAL,         /* The task is in the normal running state   */

   AOS_TASKSTATE_UNINTERRUPTIBLE,/* Un-interruptible tasks are not notified
                                  * of application resets                     */

   AOS_TASKSTATE_SUSPENDED,      /* The task is suspended                     */

   AOS_TASKSTATE_RESET_NTFY,     /* The task has been notified of an application
                                  * reset. The APPOS is waiting for an ACK.   */

   AOS_TASKSTATE_RESET_ACKED,    /* The task has ACKed the application reset
                                  * notify and is blocked, waiting to be placed
                                  * into shutdown state                       */

   AOS_TASKSTATE_SHUTDOWN,       /* The task is running in shutdown mode. It
                                  * should promptly shutdown and exit from
                                  * it's main task function.                  */

   AOS_TASKSTATE_DONE,           /* The task is suspended after completing
                                  * the task de-initialization function.
                                  * It is waiting to be re-spawned.           */

} AOS_TASKSTATE;


/* Reset detection states */
typedef enum
{
   AOS_RESETDETECT_DISABLED = 0,
   AOS_RESETDETECT_ENABLED,
   AOS_RESETDETECT_INHIBITED
} AOS_RESETDETECTMODE;


/* Reset states */
typedef enum
{
   AOS_RESETSTATE_NORESET = 0,
   AOS_RESETSTATE_RESETTING,
   AOS_RESETSTATE_POSTPONED
} AOS_RESETSTATE;


/* Synchronization event object types used by each APPOS task */
typedef enum _AOS_TASK_EVENT
{
   AOS_TASKSHUTDOWN_EVENT,    /* Tasks in the reset-ack state wait on this event
                               * until they are placed into shutdown mode.    */
   AOS_RESET_ACK_EVENT,       /* Tasks individually resetting other tasks wait
                               * on this event until the resetting task acks. */
   AOS_RESURRECT_EVENT,       /* Tasks which are "done" wait on this event until
                               * they are re-spawned. */
   AOS_RESUME_EVENT,          /* Suspended tasks wait on this event until they
                               * are resumed. */
   AOS_TASKSTART_EVENT,       /* Tasks started suspended wait on this event until
                               * they are 'started' by appOsTaskStart() */
   AOS_MAX_EVENTS
} AOS_TASK_EVENT;


/* Task control block */
typedef struct
{
   char                 name[5];          /* Task name                     */
   AOS_TASK_INIT_FNC    init;             /* Task init function            */
   AOS_TASK_MAIN_FNC    main;             /* Task main function            */
   AOS_TASK_DEINIT_FNC  deinit;           /* Task deinit function          */
   OS_TASKID            taskid;           /* Task handle                   */
   BOOL                 inuse;            /* Indicates if TCB is free      */
   AOS_TASKSTARTMODE    startMode;        /* Start SUSPENDED or RUNNING    */
   AOS_TASKSTATE        state;            /* Current task state            */
   AOS_TASKSTATE        savedState;       /* Saved task state, restored
                                           * after a suspended task is
                                           * resumed.                      */
   OS_EVENT    events[ AOS_MAX_EVENTS ];  /* Task event handles            */
   OS_TASKID            resetSrcTaskid;   /* Source of an individual reset */
} AOSZTCB;

/* Application task state */
typedef struct
{
   char                 *name;         /* App task name                 */
   int                  stackSize;     /* App task stack size           */
   AOS_APP_INIT_FNC     init;          /* App task init func            */
   AOS_APP_DEINIT_FNC   deinit;        /* App task de-init func         */
   OS_TASKARG           taskArg;       /* App task main func argument   */
   OS_TASKID            taskId;        /* App task handle               */
   OS_TASK_CLASS        taskClass;     /* App task priority             */
} AOSZAPPTASK;


/* Application OS control block */
typedef struct
{
   UINT32         nAppResets;          /* Number of application resets        */
   AOSZTCB        tcb[ MAX_TCB ];      /* TCB list                            */
   OS_SEMID       appResetSemId;       /* Wait for an app reset               */
   OS_SEMID       taskResetSemId;      /* Wait for reset ACKS from all tasks  */
   OS_SEMID       tcbMutexId;          /* Mutex to protect TCB access         */
   AOSZAPPTASK    appTask;             /* App task state                      */
   OS_TASKID      rootTaskId;          /* Root task handle                    */
   AOS_RESETSTATE resetState;          /* reset in progress/outstanding/etc   */
   AOS_RESETDETECTMODE
                  resetDetectMode;     /* Reset detection mode                */
} AOSZCB;


/*
*****************************************************************************
** LOCAL VARIABLES
*****************************************************************************
*/

PRIVATE AOSZCB aoszcb;

/*
*****************************************************************************
** LOCAL FUNCTIONS
*****************************************************************************
*/

PRIVATE OS_STATUS appOsAllocTCB( AOSZTCB **tcb );
PRIVATE OS_STATUS appOsFreeTCB( AOSZTCB *tcb );
PRIVATE OS_STATUS appOsGetTCB( OS_TASKID taskid, AOSZTCB **tcb );
PRIVATE OS_STATUS appOsGetNamedTCB( char *name, AOSZTCB **tcb );
PRIVATE char      *appOsGetTaskName( OS_TASKID taskid );
PRIVATE char      *appOsGetTaskNameSelf( void );
PRIVATE OS_STATUS appOsSetTaskState( OS_TASKID taskid, AOS_TASKSTATE state );
PRIVATE OS_STATUS appOsSetTaskStateSelf( AOS_TASKSTATE state );
PRIVATE OS_STATUS appOsSaveTaskState( OS_TASKID taskid );
PRIVATE OS_STATUS appOsGetTaskState( OS_TASKID taskid, AOS_TASKSTATE *state );
PRIVATE OS_STATUS appOsRestoreTaskState( OS_TASKID taskid );
PRIVATE BOOL      appOsResourceIsResetPending( void );
PRIVATE BOOL      appOsIsResetPendingCheck( BOOL isResource );
PRIVATE OS_STATUS appOsTaskEnterDone( void );
PRIVATE OS_STATUS appOsTaskEnterShutdown( void );
PRIVATE OS_STATUS appOsTaskResurrect( AOSZTCB *tcb );
PRIVATE OS_STATUS appOsTaskResetNotifyAll( void );
PRIVATE void      appOsRootTask( OS_TASKARG taskArg );
PRIVATE void      appOsAppTaskMain( OS_TASKARG taskArg );
PRIVATE OS_STATUS appOsSignalEvent( AOS_TASK_EVENT eventid, OS_TASKID taskid );
PRIVATE OS_STATUS appOsWaitForEvent( AOS_TASK_EVENT eventid );
PRIVATE void      appOsTaskMain( OS_TASKARG taskArg );
PRIVATE OS_STATUS appOsGetTaskStateSelf( AOS_TASKSTATE *state );
PRIVATE OS_STATUS appOsFreeSecretResources( void );

/*
** Sem create and delete macros with error checks
*/

/* Check if there was an error before creating the sem */
#define AOS_SEMCREATE( errcode, name, semid )   \
   if ( errcode == OS_STATUS_OK )               \
   {                                            \
      errcode = bcmOsSemCreate( name, semid );  \
   }

/* Check if a sem exists before deleting */
#define AOS_SEMDELETE( semid )                  \
   if ( semid )                                 \
   {                                            \
      bcmOsSemDelete( semid );                  \
   }


/*
** Mutex create and delete macros with error checks
*/

/* Check if there was an error before creating the mutex */
#define AOS_MUCREATE( errcode, name, mutexid )     \
   if ( errcode == OS_STATUS_OK )                  \
   {                                               \
      errcode = bcmOsMutexCreate( name, mutexid ); \
   }

/* Check if a mutex exists before deleting */
#define AOS_MUDELETE( mutexid )                    \
   if ( mutexid )                                  \
   {                                               \
      bcmOsMutexDelete( mutexid );                 \
   }


/*
** Event create and delete macros with error checks
*/

/* Check if there was an error before creating the event */
#define AOS_EVENTCREATE( errcode, eventid, taskid )   \
   if ( errcode == OS_STATUS_OK )                     \
   {                                                  \
      errcode = bcmOsEventCreate( eventid, taskid );  \
   }

/* Check if a event exists before deleting */
#define AOS_EVENTDELETE( eventid )                    \
   if (( eventid ))                                   \
   {                                                  \
      bcmOsEventDelete(&( eventid ));                 \
   }


/*****************************************************************************
** FUNCTION:   appOsInit
**
** PURPOSE:    Initialize any OS specific variables
**
** PARAMETERS: priorities - structure that contains the absolute priority
**                          levels associated with each class of MTA task.
**
** RETURNS:    OS_STATUS_OK or error code
**
** NOTE:
*****************************************************************************/
PUBLIC OS_STATUS appOsInit( OS_THREAD_CLASS_PRIO *mtaPriorities )
{
   OS_STATUS retval;

   LOG(( "APPOS: Enter appOsInit " ));

   /*
   ** Initialize control block
   */
   memset( &aoszcb, 0, sizeof( aoszcb ) );
   aoszcb.resetState = AOS_RESETSTATE_NORESET;

   /*
   ** Initialize the MTA OS abstraction layer.
   ** NOTE: This should be done before any calls are made to the OS module.
   */
   retval = bcmOsInit( mtaPriorities );
   if ( retval != OS_STATUS_OK )
   {
      LOG(( "ERROR: bcmOsInit() failed\n" ));
   }

   /*
   ** Create appOS semaphores & mutexes
   */
   AOS_SEMCREATE( retval, AOS_APPRESET_SEM,  &aoszcb.appResetSemId   );
   AOS_SEMCREATE( retval, AOS_TASKRESET_SEM, &aoszcb.taskResetSemId  );
   AOS_MUCREATE(  retval, AOS_TCB_MUTEX,     &aoszcb.tcbMutexId      );      

   if ( retval != OS_STATUS_OK )
   {
      LOG(( "appOsInit: Initialization error, cleaning up semaphores/mutexes" ));

      /* Cleanup anything created before returning */
      AOS_SEMDELETE( aoszcb.appResetSemId    );
      AOS_SEMDELETE( aoszcb.taskResetSemId   );
      AOS_MUDELETE(  aoszcb.tcbMutexId       );

      return ( retval );
   }

   LOG(( "APPOS: Exit appOsInit " ));
   return( retval );
}


/*****************************************************************************
** FUNCTION:   appOsTaskMain
**
** PURPOSE:    Task main template for all application tasks
**
** PARAMETERS: taskArg - argument passed to main task function
**
** RETURNS:    none
**
** NOTE:
*****************************************************************************/
PRIVATE void appOsTaskMain( OS_TASKARG taskArg )
{
   OS_STATUS            retval;
   OS_TASKID            taskid;
   AOSZTCB              *t;
   AOS_TASK_INIT_FNC    init;
   AOS_TASK_MAIN_FNC    main;
   AOS_TASK_DEINIT_FNC  deinit;
   AOS_TASKSTARTMODE    taskStartMode;


   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   retval = bcmOsTaskIdentSelf( &taskid );
   if ( retval != OS_STATUS_OK )
   {
      LOG(( "HUGE trouble in little China!!! Couldn't start appOsTaskMain!" ));
      return;
   }

   retval = appOsGetTCB( taskid, &t );
   if ( retval != OS_STATUS_OK )
   {
      /* Leave critical section */
      bcmOsMutexUnlock( aoszcb.tcbMutexId );

      LOG(( "HUGE trouble in little China!!! Couldn't start appOsTaskMain!" ));
      return;
   }

   init = t->init;
   main = t->main;
   deinit = t->deinit;
   taskStartMode = t->startMode;

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );


   /* Invoke task's init, main, and deinit functions, looping forever */
   while ( 1 )
   {
      if ( taskStartMode == AOS_TASKSTART_SUSPENDED )
      {
         LOG(( "APPOS: appOsTaskMain() task %s started suspended", appOsGetTaskNameSelf() ));
         appOsWaitForEvent( AOS_TASKSTART_EVENT );
         LOG(( "APPOS: appOsTaskMain() task %s started (from suspended)", appOsGetTaskNameSelf() ));
      }
      (*init)();
      (*main)( taskArg );
      appOsTaskEnterShutdown();
      (*deinit)();
      appOsTaskEnterDone();
   }
}


/*****************************************************************************
** FUNCTION:   appOsTaskCreateMode
**
** PURPOSE:    Create a task
**
** PARAMETERS: name - name of the task (4 characters max)
**             stackSize - task stack size (bytes)
**             taskClass - task priority class
**             taskInit - task initialization function
**             taskMain - task main function
**             taskDeinit - task de-initialization function
**             argument - the task argument passed to <taskMain>
**             taskId - task identifier
**             bUnInterruptible - indicates if task is uninterruptible.
**                      Uninterruptible tasks do not allocate or use OS resources,
**                      and therefore are not informed of application resets.
**             taskStartMode - Mode to start the task (AOS_TASKSTARTMODE_RUNNING | SUSPENDED)
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
******************************************************************************/
PUBLIC OS_STATUS appOsTaskCreateMode( char* name,
                                      int stackSize,
                                      OS_TASK_CLASS taskClass,
                                      AOS_TASK_INIT_FNC taskInit,
                                      AOS_TASK_MAIN_FNC taskMain,
                                      AOS_TASK_DEINIT_FNC taskDeinit,
                                      OS_TASKARG argument,
                                      OS_TASKID *taskId,
                                      BOOL bUnInterruptible,
                                      AOS_TASKSTARTMODE taskStartMode )
{
   AOSZTCB     *tcb;
   OS_STATUS   retval;

   LOG(( "APPOS: Enter appOsTaskCreate " ));

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );


   /* Creation of new tasks is not allowed during application reset since
    * the newly created task will never be notified of the reset
    * (so task creation is only allowed during NORESET or POSTPONED). */
   if ( aoszcb.resetState == AOS_RESETSTATE_RESETTING )
   {
      LOG(( "APPOS: Can't create task (%s) during a reset!", name ));
      bcmOsMutexUnlock( aoszcb.tcbMutexId );
      return ( AOS_STATUS_RESET );
   }

   /* Check if we are resurrecting a task that was shutdown */
   retval = appOsGetNamedTCB( name, &tcb );
   if ( retval == OS_STATUS_OK )
   {
      /*
      ** Ressurect this task if:
      ** we found the tcb, state is shutdown, and it matches the args passed
      */
      if (     ( retval          == OS_STATUS_OK            )
            && ( tcb->state      == AOS_TASKSTATE_DONE      )
            && ( tcb->init       == taskInit                )
            && ( tcb->main       == taskMain                )
            && ( tcb->deinit     == taskDeinit              )
            && ( tcb->startMode  == taskStartMode           )
         )
      {
         LOG(( "APPOS: appOsTaskCreate - taskRessurrect %s", name ));

         retval = appOsTaskResurrect( tcb );
         if ( retval == OS_STATUS_OK )
         {
            /* Set the caller's task id */
            *taskId = tcb->taskid;
         }

         /* Leave critical section */
         bcmOsMutexUnlock( aoszcb.tcbMutexId );
         return ( retval );
      }

      LOG(( "ERROR: TCB mismatch when resurrecting task %s (possible duplicate task name).", name ));
      ASSERT( 0 );
      /* return ( OS_STATUS_ERR ); */
   }


   /*
   ** Create the task from scratch
   */
   retval = appOsAllocTCB( &tcb );
   if ( retval != OS_STATUS_OK )
   {
      bcmOsMutexUnlock( aoszcb.tcbMutexId );
      return ( retval );
   }

   strncpy( tcb->name, name, 4 );
   tcb->name[4] = '\0';
   tcb->init = taskInit;
   tcb->main = taskMain;
   tcb->deinit = taskDeinit;
   tcb->startMode = taskStartMode;

   /* Need to determine if this is an Un-interruptible task. */
   if ( bUnInterruptible )
   {
      tcb->state = AOS_TASKSTATE_UNINTERRUPTIBLE;
   }
   else
   {
      tcb->state = AOS_TASKSTATE_NORMAL;
   }

   /*
   ** Spawn the task.
   */
   LOG(( "APPOS: appOsTaskCreate - spawn new task %s", name ));

   /* Pass in tcb.taskid since it must be valid as soon as the task is created
    *  because it will be used by the task to check it's own state */
   retval = bcmOsTaskCreate( name, stackSize, taskClass, appOsTaskMain, argument, &tcb->taskid );

   /*
   ** Create tasks's event synchronization objects
   */
   AOS_EVENTCREATE( retval, &tcb->events[ AOS_TASKSHUTDOWN_EVENT ],  &tcb->taskid );
   AOS_EVENTCREATE( retval, &tcb->events[ AOS_RESURRECT_EVENT ],     &tcb->taskid );
   AOS_EVENTCREATE( retval, &tcb->events[ AOS_RESUME_EVENT ],        &tcb->taskid );
   AOS_EVENTCREATE( retval, &tcb->events[ AOS_TASKSTART_EVENT ],     &tcb->taskid );
   AOS_EVENTCREATE( retval, &tcb->events[ AOS_RESET_ACK_EVENT ],     &tcb->taskid );

   if ( retval != OS_STATUS_OK )
   {
      LOG(( "appOsTaskCreate: Initialization error, cleaning up task state" ));

      /* Cleanup anything created before returning */
      AOS_EVENTDELETE( tcb->events[ AOS_TASKSHUTDOWN_EVENT ]   );
      AOS_EVENTDELETE( tcb->events[ AOS_RESURRECT_EVENT ]      );
      AOS_EVENTDELETE( tcb->events[ AOS_RESUME_EVENT ]         );
      AOS_EVENTDELETE( tcb->events[ AOS_TASKSTART_EVENT ]      );
      AOS_EVENTDELETE( tcb->events[ AOS_RESET_ACK_EVENT ]      );

      appOsFreeTCB( tcb );
   }
   else
   {
      /* If the create succeeded, set the callers taskid */
      *taskId = tcb->taskid;
   }

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   LOG(( "APPOS: Exit appOsTaskCreate " ));
   return ( retval );
}


/*****************************************************************************
** FUNCTION:   appOsTaskDelete
**
** PURPOSE:    Deletes a task
**
** PARAMETERS: taskId - task identification
**
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
******************************************************************************/
PUBLIC OS_STATUS appOsTaskDelete( OS_TASKID taskId )
{
   AOSZTCB      *tcb;
   OS_STATUS   status;

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   /* Self-identify ( taskid is zero ) */
   if ( taskId == 0 )
   {
      status = bcmOsTaskIdentSelf( &taskId );
      if ( status != OS_STATUS_OK )
      {
         bcmOsMutexUnlock( aoszcb.tcbMutexId );
         return ( status );
      }
   }

   status = appOsGetTCB( taskId, &tcb );
   if ( status != OS_STATUS_OK )
   {
      bcmOsMutexUnlock( aoszcb.tcbMutexId );
      return ( status );
   }

   /*
   ** Free the local storage associated with the task
   */

   AOS_EVENTDELETE( tcb->events[ AOS_TASKSHUTDOWN_EVENT ]   );
   AOS_EVENTDELETE( tcb->events[ AOS_RESURRECT_EVENT ]      );
   AOS_EVENTDELETE( tcb->events[ AOS_RESUME_EVENT ]         );
   AOS_EVENTDELETE( tcb->events[ AOS_TASKSTART_EVENT ]      );
   AOS_EVENTDELETE( tcb->events[ AOS_RESET_ACK_EVENT ]      );

   appOsFreeTCB( tcb );

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   /* Delete the task */
   return ( bcmOsTaskDelete( taskId ));
}


/*****************************************************************************
** FUNCTION:   appOsTaskStart
**
** PURPOSE:    Start a task which was created in suspended mode
**
** PARAMETERS: taskId - task identifier
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
******************************************************************************/
PUBLIC OS_STATUS appOsTaskStart( OS_TASKID taskId )
{
   return ( appOsSignalEvent( AOS_TASKSTART_EVENT, taskId ));
}


/*****************************************************************************
** FUNCTION:   appOsSleep
**
** PURPOSE:    Sleep for a specified number of ticks
**
** PARAMETERS: sleeptime - the time to sleep for (in ticks)
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PUBLIC OS_STATUS appOsSleep( OS_TICKS sleeptime )
{
   OS_STATUS retval;

   while ( !appOsTaskIsResetPending() )
   {
      if ( sleeptime <= RESET_GRANULARITY_TICKS )
      {
         /* Sleeptime is less than the reset granularity, simply sleep normally */
         return ( bcmOsSleep( sleeptime ) );
      }
      else
      {
         /* Sleeptime is more than reset granularity, decrement sleeptime remaining
         ** and sleep for the maximum allowed before checking for a reset */
         if ( sleeptime != OS_WAIT_FOREVER )
         {
            sleeptime -= RESET_GRANULARITY_TICKS;
         }
         retval = bcmOsSleep( RESET_GRANULARITY_TICKS );
         if ( retval != OS_STATUS_OK )
         {
            /* Anything but the OK to this chunk of the sleep is returned to the caller */
            return (retval);
         }
      }
   }

   return ( AOS_STATUS_RESET );
}


/*****************************************************************************
** FUNCTION:   appOsTaskSuspend
**
** PURPOSE:    Suspends a task
**
** PARAMETERS: taskId - task identifier  ( 0 for suicide )
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
******************************************************************************/
PUBLIC OS_STATUS appOsTaskSuspend( OS_TASKID taskId )
{
   OS_STATUS retval;

   /* Only self-suspension is currently supported */
   if ( taskId != 0 )
   {
      return ( OS_STATUS_ERR );
   }

   if ( appOsTaskIsResetPending() )
   {
      /* Reset pending, return immediately */
      return ( AOS_STATUS_RESET );
   }

   /*
   ** No pending reset, suspend the task
   */

   /* Self-identify ( taskid is zero ) */
   retval = bcmOsTaskIdentSelf( &taskId );
   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   /* Save task state to be restored after suspension */
   appOsSaveTaskState( taskId );

   /* Set new task state */
   appOsSetTaskState( taskId, AOS_TASKSTATE_SUSPENDED );

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   /* Suspend task */
   retval = appOsWaitForEvent( AOS_RESUME_EVENT );
   if ( retval != OS_STATUS_OK )
   {
      /* Enter critical section */
      bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

      /* Suspend failed, revert state and return error */
      appOsRestoreTaskState( taskId );

      /* Leave critical section */
      bcmOsMutexUnlock( aoszcb.tcbMutexId );
      return ( retval );
   }

   /* Suspend succeeded, check for pending reset after being resumed */
   if ( appOsTaskIsResetPending() )
   {
      /* Reset pending, return RESET */
      retval = AOS_STATUS_RESET;
   }

   return ( retval );
}
          

/*****************************************************************************
** FUNCTION:   appOsTaskResume
**
** PURPOSE:    Resumes a suspended a task
**
** PARAMETERS: taskId - task identifier
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
******************************************************************************/
PUBLIC OS_STATUS appOsTaskResume( OS_TASKID taskId )
{
   OS_STATUS retval;
   AOS_TASKSTATE state;

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   /* Verify that the task is currently suspended */
   retval = appOsGetTaskState( taskId, &state );
   if ( retval != OS_STATUS_OK )
   {
      bcmOsMutexUnlock( aoszcb.tcbMutexId );
      return ( retval );
   }

   if ( state != AOS_TASKSTATE_SUSPENDED )
   {
      LOG(( "ERROR: appOsTaskResume() attempting to resume a task not in SUSPENDED state!" ));
      bcmOsMutexUnlock( aoszcb.tcbMutexId );
      return ( OS_STATUS_ERR );
   }

   retval = appOsRestoreTaskState( taskId );

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   /* Resume task */
   return ( appOsSignalEvent( AOS_RESUME_EVENT, taskId ));
}


/*****************************************************************************
** FUNCTION:   appOsSemTake
**
** PURPOSE:    Take a semaphore
**
** PARAMETERS: semId - Handle of the semaphore
**             timeout_ticks - timeout in ticks
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
******************************************************************************/
PUBLIC OS_STATUS appOsSemTake(OS_SEMID semId, OS_TICKS timeout)
{
   OS_STATUS retval;

   while ( !appOsResourceIsResetPending() )
   {
      if ( timeout <= RESET_GRANULARITY_TICKS )
      {
         /* Timeout is less than the reset granularity, semTake normally */
         return ( bcmOsSemTake( semId, timeout ) );
      }
      else
      {
         /* Timeout is more than reset granularity, decrement timeout remaining
         ** and timeout for the maximum allowed before checking for a reset */
         if ( timeout != OS_WAIT_FOREVER )
         {
            timeout -= RESET_GRANULARITY_TICKS;
         }
         retval = bcmOsSemTake( semId, RESET_GRANULARITY_TICKS );
         if ( retval != OS_STATUS_TIMEOUT )
         {
            /* Anything but the early timeout is returned to the caller */
            return (retval);
         }
      }
   }

   return ( AOS_STATUS_RESET );
}


/*****************************************************************************
** FUNCTION:   appOsMutexLock
**
** PURPOSE:    Lock a mutex
**
** PARAMETERS: mutexId        - Handle of the mutex
**             timeout_ticks  - timeout in ticks
**
** RETURNS:    OS_STATUS_OK   - success
**             error code     - failure
**
** NOTE:
******************************************************************************/
PUBLIC OS_STATUS appOsMutexLock(OS_MUTEXID mutexId, OS_TICKS timeout)
{
   OS_STATUS retval;

   while ( !appOsResourceIsResetPending() )
   {
      if ( timeout <= RESET_GRANULARITY_TICKS )
      {
         /* Timeout is less than the reset granularity, MutexLock normally */
         return ( bcmOsMutexLock( mutexId, timeout ) );
      }
      else
      {
         /* Timeout is more than reset granularity, decrement timeout remaining
         ** and timeout for the maximum allowed before checking for a reset */
         if ( timeout != OS_WAIT_FOREVER )
         {
            timeout -= RESET_GRANULARITY_TICKS;
         }
         retval = bcmOsMutexLock( mutexId, RESET_GRANULARITY_TICKS );
         if ( retval != OS_STATUS_TIMEOUT )
         {
            /* Anything but the early timeout is returned to the caller */
            return (retval);
         }
      }
   }

   return ( AOS_STATUS_RESET );
}


/*****************************************************************************
** FUNCTION:   appOsQueueReceive
**
** PURPOSE:    Receives a message from the specified message queue
**
** PARAMETERS: qid      - Message Queue ID
**             timeout  - Timeout in clock ticks
**             msgBuf   - Message Buffer
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       If timeout is 0, bcmOsQueueReceive() will be non-blocking,
**             i.e. it will return even if there is no available message
**             in the queue. If timeout is not 0, then this function will
**             block until a message is posted to the queue or the timeout
**             occurs, which ever comes first.
**
******************************************************************************/
PUBLIC OS_STATUS appOsQueueReceive( OS_QID qid, OS_TICKS timeout, OS_QMSG *qMsg )
{
   OS_STATUS retval;

   while ( !appOsResourceIsResetPending() )
   {
      if ( timeout <= RESET_GRANULARITY_TICKS )
      {
         /* Timeout is less than the reset granularity, queueReceive normally */
         return ( bcmOsQueueReceive( qid, timeout, qMsg ));
      }
      else
      {
         /* Timeout is more than reset granularity, decrement timeout remaining
         ** and timeout for the maximum allowed before checking for a reset */
         if ( timeout != OS_WAIT_FOREVER )
         {
            timeout -= RESET_GRANULARITY_TICKS;
         }
         retval = bcmOsQueueReceive( qid, RESET_GRANULARITY_TICKS, qMsg );
         if ( retval != OS_STATUS_TIMEOUT )
         {
            /* Anything but the early timeout is returned to the caller */
            return (retval);
         }
      }
   }

   return ( AOS_STATUS_RESET );
}


/*****************************************************************************
** FUNCTION:   appOsSelect
**
** PURPOSE:    This function is used to multiplex I/O requests among multiple
**             sockets. Three sets of socket descriptors may be specified: a
**             set of sockets from which to read, a set to which to write and
**             a set that may have pending exceptional conditions.
**
**             If bcmOsSelect returns successfully, the three sets indicate
**             which socket descriptors can be read, which can be written to,
**             or which have exceptional conditions pending. A timeout value
**             may be specified.
**
** PARAMETERS: readset     [IN] - Set of sockets from which to read
**             writeset    [IN] - Set of sockets from which to write
**             exceptset   [IN] - Set of sockets that may have an exceptional
**                                condition pending
**             timeout     [IN] - Specifies a timeout option. If the timeout is
**                                set to 0, bcmOsSelect() returns immediately. If
**                                the timeout is OS_WAIT_FOREVER, bcmOsSelect()
**                                blocks until a descriptor is selectable.
**             numDesc     [OUT] - On success a nonzero value is returned that
**                                 indicates the number of descriptors on which
**                                 events have occurred.
**
** RETURNS:   OS_STATUS_OK or error code
**
** NOTE:
*****************************************************************************/
OS_STATUS appOsSelect( OS_FD_SET readset,
                       OS_FD_SET writeset,
                       OS_FD_SET exceptset,
                       OS_TICKS timeout,
                       int *numDesc )
{
   OS_STATUS   retval;
   OS_FD_SET   origRdSet;
   OS_FD_SET   origWrSet;
   OS_FD_SET   origExSet;
   BOOL        bRdSet;
   BOOL        bWrSet;
   BOOL        bExSet;

   bRdSet = ( readset != OS_FD_SET_NULL );
   bWrSet = ( writeset != OS_FD_SET_NULL );
   bExSet = ( exceptset != OS_FD_SET_NULL );

   /*
   ** Open new fdsets to save original aruments
   */
   if ( bRdSet )
   {
      retval = bcmOsFdSetOpen( &origRdSet );
      if ( retval != OS_STATUS_OK )
      {
         bcmOsFdSetClose( origRdSet );
         return ( retval );
      }
   }

   if ( bWrSet )
   {
      retval = bcmOsFdSetOpen( &origWrSet );
      if ( retval != OS_STATUS_OK )
      {
         if ( bRdSet ) bcmOsFdSetClose( origRdSet );
         bcmOsFdSetClose( origWrSet );
         return ( retval );
      }
   }

   if ( bExSet )
   {
      retval = bcmOsFdSetOpen( &origExSet );
      if ( retval != OS_STATUS_OK )
      {
         if ( bRdSet ) bcmOsFdSetClose( origRdSet );
         if ( bWrSet ) bcmOsFdSetClose( origWrSet );
         bcmOsFdSetClose( origExSet );
         return ( retval );
      }
   }

   /*
   ** Save the original fdsets to restore after multiple select() calls
   */
   if ( bRdSet ) bcmOsFdSetCopy( origRdSet, readset );
   if ( bWrSet ) bcmOsFdSetCopy( origWrSet, writeset );
   if ( bExSet ) bcmOsFdSetCopy( origExSet, exceptset );

   /*
   ** Repeat the select loop until the true timeout expires, a reset occurs, or select succeeds
   */
   retval = AOS_STATUS_RESET;
   while ( !appOsResourceIsResetPending() )
   {
      /* Restore the original sets */
      if ( bRdSet ) bcmOsFdSetCopy( readset, origRdSet );
      if ( bWrSet ) bcmOsFdSetCopy( writeset, origWrSet );
      if ( bExSet ) bcmOsFdSetCopy( exceptset, origExSet );

      if ( timeout <= RESET_GRANULARITY_TICKS )
      {
         /* timeout is less than the reset granularity, select normally */
         retval = bcmOsSelect( readset, writeset, exceptset, timeout, numDesc );
         break;
      }
      else
      {
         /* Timeout is more than reset granularity, decrement timeout remaining
         ** and timeout for the maximum allowed before checking for a reset */
         if ( timeout != OS_WAIT_FOREVER )
         {
            timeout -= RESET_GRANULARITY_TICKS;
         }
         retval = bcmOsSelect( readset, writeset, exceptset, RESET_GRANULARITY_TICKS, numDesc );
         if ( retval != OS_STATUS_TIMEOUT )
         {
            /* Anything but the early timeout is returned to the caller */
            break;
         }
      }

      retval = AOS_STATUS_RESET;
   }

   /* Cleanup and return */
   if ( bRdSet ) bcmOsFdSetClose( origRdSet );
   if ( bWrSet ) bcmOsFdSetClose( origWrSet );
   if ( bExSet ) bcmOsFdSetClose( origExSet );

   return ( retval );
}


/*****************************************************************************
** FUNCTION:   appOsSocketReceiveFrom
**
** PURPOSE:    Receive a message from a socket
**
** PARAMETERS:
**             sd            - socket to receive from
**             buf           - pointer to data buffer
**             buflen        - length of buffer
**             flags         - flags to underlying protocols
**             ifno           - specifies outgoing interface for message
**                               (0 for default interface)
**             from          - where to copy sender's address
**             fromLen       - pointer to size of "from" buffer
**             numBytesRec   - number of bytes received
**
** RETURNS:    OS_STATUS_OK  - success
**             error code    - failure
**
** NOTE:
******************************************************************************/
OS_STATUS appOsSocketReceiveFrom( int           sd,
                                  char          *buf,
                                  int           buflen,
                                  int           flags,
                                  long          ifno,
                                  OS_SOCKADDR   *from,
                                  int           *fromLen,
                                  int           *numBytesRec)
{
   OS_FD_SET   fdSet;
   OS_STATUS   status;
   int         nDesc;

   /* Create a file descriptor set */
   status = bcmOsFdSetOpen( &fdSet );
   if ( status != OS_STATUS_OK )
   {
      return ( status );
   }

   /* Clear the file descriptor set */
   status = bcmOsFdZero( fdSet );
   if ( status != OS_STATUS_OK )
   {
      bcmOsFdSetClose( fdSet );
      return ( status );
   }

   /* Add receive socket to the file descriptor set */
   status = bcmOsFdSet( sd, fdSet );
   if ( status != OS_STATUS_OK )
   {
      bcmOsFdSetClose( fdSet );
      return ( status );
   }

   /* Block until data is ready to be received from the socket */
   status = appOsSelect( fdSet, OS_FD_SET_NULL, OS_FD_SET_NULL, OS_WAIT_FOREVER, &nDesc );
   bcmOsFdSetClose( fdSet );
   if ( status != OS_STATUS_OK )
   {
      return ( status );
   }

   /* Read received data from the socket */
   return ( bcmOsSocketReceiveFrom( sd, buf, buflen, flags,
                                    ifno, from, fromLen, numBytesRec ));
}


/*****************************************************************************
** FUNCTION:   appOsTftpReadFile
**
** PURPOSE:    This function reads the specified file from the specified TFTP
**             server, placing the file data in the buffer.
**             Up to "bufferSize" bytes of the file will be read.
**
** PARAMETERS: serverIpAddress (I)  - the IP address of the TFTP server.
**             pFilename       (I)  - the fully qualified name of the file
**                                    to be read.
**             ifno            (I)  - IP Iface # over which the TFTP is performed
**             pBuffer         (O)  - buffer with retrieved data.
**             bufferSize      (I)  - Size of the buffer.
**             pBytesRead      (O)  - the number of bytes actually read.
**             abortFunc       (I)  - function to call to check if file read should
**                                    be aborted mid-stream.
**
** RETURNS:    OS_STATUS_OK   - success
**             error code     - failure
**
** NOTE:       appOs version ignores the abortFunc argument and always uses
**             appOsTaskIsResetPending() as the abortFunc
*****************************************************************************/
OS_STATUS appOsTftpReadFile(  IPADDR serverIpAddress,
                              char *pFilename,
                              UINT32 ifno,
                              char *pBuffer,
                              UINT32 bufferSize,
                              UINT32 *pBytesRead,
                              OS_ABORT_FNC abortFunc )
{
   /* Call bcmOs version of TftpRead w/IsResetPending() as the abort handler */
   return ( bcmOsTftpReadFile(   serverIpAddress, pFilename, ifno,
                                 pBuffer, bufferSize, pBytesRead,
                                 appOsTaskIsResetPending )
          );
}


/*****************************************************************************
** FUNCTION:   appOsEventReceive
**
** PURPOSE:    Wait for an event to be sent to the calling task
**
** PARAMETERS: event - event to wait for
**             timeout - event timeout delay (in ticks)
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
******************************************************************************/
OS_STATUS appOsEventReceive( OS_EVENT *event, OS_TICKS timeout )
{
   OS_STATUS retval;

   while ( !appOsResourceIsResetPending() )
   {
      if ( timeout <= RESET_GRANULARITY_TICKS )
      {
         /* Timeout is less than the reset granularity, event-receive normally */
         return ( bcmOsEventReceive( event, timeout ));
      }
      else
      {
         /* Timeout is more than reset granularity, decrement timeout remaining
         ** and timeout for the maximum allowed before checking for a reset */
         if ( timeout != OS_WAIT_FOREVER )
         {
            timeout -= RESET_GRANULARITY_TICKS;
         }
         retval = bcmOsEventReceive( event, RESET_GRANULARITY_TICKS );
         if ( retval != OS_STATUS_TIMEOUT )
         {
            /* Anything but the early timeout is returned to the caller */
            return (retval);
         }
      }
   }

   return ( AOS_STATUS_RESET );
}


/*****************************************************************************
** FUNCTION:   appOsAllocTCB
**
** PURPOSE:    Allocate a new task control block
**
**             tcb - returned pointer to the allocated TCB
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       It is the caller's responsibility to insure concurrency protection
*****************************************************************************/
PRIVATE OS_STATUS appOsAllocTCB( AOSZTCB **tcb )
{
   AOSZTCB *t;
   int i;

   for ( i = 0; i < MAX_TCB; i++ )
   {
      t = &aoszcb.tcb[i];

      if ( !t->inuse )
      {
         t->inuse = 1;
         *tcb = t;
         return ( OS_STATUS_OK );
      }
   }

   *tcb = NULL;
   return ( OS_STATUS_ERR );
}


/*****************************************************************************
** FUNCTION:   appOsFreeTCB
**
** PURPOSE:    De-allocate a task control block
**
** PARAMETERS: tcb - pointer to TCB to free
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:      It is the caller's responsibility to insure concurrency protection
*****************************************************************************/
PRIVATE OS_STATUS appOsFreeTCB( AOSZTCB *tcb )
{
   memset( tcb, 0, sizeof( AOSZTCB ));

   return ( OS_STATUS_OK );

}


/*****************************************************************************
** FUNCTION:   appOsGetTCB
**
** PURPOSE:    To retrieve a task control block associated with a task id
**
** PARAMETERS: taskid - task identifier
**             tcb - returned pointer to the TCB
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       It is the caller's responsibility to insure concurrency protection
*****************************************************************************/
PRIVATE OS_STATUS appOsGetTCB( OS_TASKID taskid, AOSZTCB **tcb )
{
   AOSZTCB *t;
   int i;

   for ( i = 0; i < MAX_TCB; i++ )
   {
      t = &aoszcb.tcb[i];

      if ( t->inuse && ( t->taskid == taskid ) )
      {
         *tcb = t;
         return ( OS_STATUS_OK );
      }
   }

   *tcb = NULL;
   return ( OS_STATUS_ERR );
}


/*****************************************************************************
** FUNCTION:   appOsGetNamedTCB
**
** PURPOSE:    To retrieve the task control block associated with a task name
**
** PARAMETERS: name - task name
**             tcb - returned pointer to the TCB
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       It is the caller's responsibility to insure concurrency protection
*****************************************************************************/
PRIVATE OS_STATUS appOsGetNamedTCB( char *name, AOSZTCB **tcb )
{
   AOSZTCB *t;
   int i;

   for ( i = 0; i < MAX_TCB; i++ )
   {
      t = &aoszcb.tcb[i];

      /* Max TCB name length is 4 characters */
      if ( t->inuse && !strncmp( t->name, name, 4 ))
      {
         *tcb = t;
         return ( OS_STATUS_OK );
      }
   }

   *tcb = NULL;
   return ( OS_STATUS_ERR );
}


/*****************************************************************************
** FUNCTION:   appOsGetTaskName
**
** PURPOSE:    Get the name of specified task
**
** PARAMETERS: taskid - task identifier
**
** RETURNS:    name or "NULL"
**
** NOTE:
*****************************************************************************/
PRIVATE char *appOsGetTaskName( OS_TASKID taskid )
{
   OS_STATUS   retval;
   AOSZTCB     *tcb;

   retval = appOsGetTCB( taskid, &tcb );
   if ( retval == OS_STATUS_OK )
   {
      return ( tcb->name );
   }
   else
   {
      return ( "NULL" );
   }
}


/*****************************************************************************
** FUNCTION:   appOsGetTaskNameSelf
**
** PURPOSE:    Get the name of the calling task
**
** PARAMETERS: none
**
** RETURNS:    name or "NULL"
**
** NOTE:
*****************************************************************************/
PRIVATE char *appOsGetTaskNameSelf( void )
{
   OS_TASKID   taskid;

   bcmOsTaskIdentSelf( &taskid );

   return ( appOsGetTaskName( taskid ) );
}


/*****************************************************************************
** FUNCTION:   appOsWaitForEvent
**
** PURPOSE:    The calling task will wait for the specified event to occur
**
** PARAMETERS: eventid - the event type to wait for
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PRIVATE OS_STATUS appOsWaitForEvent( AOS_TASK_EVENT eventid )
{
   OS_TASKID   taskid;
   AOSZTCB     *tcb;
   OS_STATUS   retval;

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   /* Get my TCB */
   retval = bcmOsTaskIdentSelf( &taskid );
   if ( retval == OS_STATUS_OK )
   {
      retval = appOsGetTCB( taskid, &tcb );
   }

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   if ( retval == OS_STATUS_OK )
   {
      /* Wait until the event is received */
      retval = bcmOsEventReceive( &tcb->events[ eventid ], OS_WAIT_FOREVER );
   }

   return ( retval );
}


/*****************************************************************************
** FUNCTION:   appOsSignalEvent
**
** PURPOSE:    To send an event to a task
**
** PARAMETERS: eventid - the event type to send
**             taskid - the task to send the event to
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PRIVATE OS_STATUS appOsSignalEvent( AOS_TASK_EVENT eventid, OS_TASKID taskid )
{
   AOSZTCB     *tcb;
   OS_STATUS   retval;

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   retval = appOsGetTCB( taskid, &tcb );

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   if ( retval == OS_STATUS_OK )
   {
      /* Send an event to the specified task */
      retval = bcmOsEventSend( &tcb->events[ eventid ] );
   }

   return ( retval );
}


/*****************************************************************************
** FUNCTION:   appOsSetTaskState
**
** PURPOSE:    Set the task state of the specified task
**
** PARAMETERS: taskid - task id of the task to change the state of
**             state  - new task state for the specified task
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       It is the caller's responsibility to insure concurrency protection
*****************************************************************************/
PRIVATE OS_STATUS appOsSetTaskState( OS_TASKID taskid, AOS_TASKSTATE state )
{
   OS_STATUS retval;
   AOSZTCB *tcb;

   retval = appOsGetTCB( taskid, &tcb );

   if ( retval == OS_STATUS_OK )
   {
      tcb->state = state;
   }

   return ( retval );
}


/*****************************************************************************
** FUNCTION:   appOsSetTaskStateSelf
**
** PURPOSE:    Set the task state of the calling task
**
** PARAMETERS: state  - new task state for the calling task
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:      It is the caller's responsibility to insure concurrency protection
*****************************************************************************/
PRIVATE OS_STATUS appOsSetTaskStateSelf( AOS_TASKSTATE state )
{
   OS_STATUS retval;
   OS_TASKID taskid;

   retval = bcmOsTaskIdentSelf( &taskid );

   if ( retval == OS_STATUS_OK )
   {
      retval = appOsSetTaskState( taskid, state );
   }

   return ( retval );
}


/*****************************************************************************
** FUNCTION:   appOsGetTaskState
**
** PURPOSE:    Get the task state of the specified task
**
** PARAMETERS: taskid - get state of this task
**             state  - current state for the specified task
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       It is the caller's responsibility to insure concurrency protection
*****************************************************************************/
PRIVATE OS_STATUS appOsGetTaskState( OS_TASKID taskid, AOS_TASKSTATE *state )
{
   OS_STATUS retval;
   AOSZTCB *tcb;

   if ( !state )
   {
      return ( OS_STATUS_ERR );
   }

   retval = appOsGetTCB( taskid, &tcb );

   if ( retval == OS_STATUS_OK )
   {
      *state = tcb->state;
   }

   return ( retval );
}


/*****************************************************************************
** FUNCTION:   appOsGetTaskStateSelf
**
** PURPOSE:    Get the task state of the calling task
**
** PARAMETERS: state  - current state for the calling task
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PRIVATE OS_STATUS appOsGetTaskStateSelf( AOS_TASKSTATE *state )
{
   OS_STATUS retval;
   OS_TASKID taskid;

   retval = bcmOsTaskIdentSelf( &taskid );

   if ( retval == OS_STATUS_OK )
   {
      /* Enter critical section */
      bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

      retval = appOsGetTaskState( taskid, state );

      /* Leave critical section */
      bcmOsMutexUnlock( aoszcb.tcbMutexId );
   }

   return ( retval );
}


/*****************************************************************************
** FUNCTION:   appOsSaveTaskState
**
** PURPOSE:    Save the state of a task (restored later by appOsRestoreTaskState)
**
** PARAMETERS: taskid - task id of the task to save the state
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       It is the caller's responsibility to insure concurrency protection
*****************************************************************************/
PRIVATE OS_STATUS appOsSaveTaskState( OS_TASKID taskid )
{
   OS_STATUS retval;
   AOSZTCB *tcb;

   retval = appOsGetTCB( taskid, &tcb );

   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   return ( appOsGetTaskState( taskid, &tcb->savedState ) );
}


/*****************************************************************************
** FUNCTION:   appOsRestoreTaskState
**
** PURPOSE:    Restore the state of a task (saved by appOsSaveTaskState)
**
** PARAMETERS: taskid - task id of the task to restore the state
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       It is the caller's responsibility to insure concurrency protection
*****************************************************************************/
PRIVATE OS_STATUS appOsRestoreTaskState( OS_TASKID taskid )
{
   OS_STATUS retval;
   AOSZTCB *tcb;

   retval = appOsGetTCB( taskid, &tcb );

   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   return ( appOsSetTaskState( taskid, tcb->savedState ) );
}


/*****************************************************************************
** FUNCTION:   appOsTaskIsResetPending
**
** PURPOSE:    Checks if the calling task has been notified of a pending reset
**
** PARAMETERS:
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PUBLIC BOOL appOsTaskIsResetPending( void )
{
   return ( appOsIsResetPendingCheck( FALSE ) );
}


/*****************************************************************************
** FUNCTION:   appOsResourceIsResetPending
**
** PURPOSE:    Checks if the calling task has been notified of a pending reset
**
** PARAMETERS:
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PUBLIC BOOL appOsResourceIsResetPending( void )
{
   return ( appOsIsResetPendingCheck( TRUE ) );
}


/*****************************************************************************
** FUNCTION:   appOsIsResetPending
**
** PURPOSE:    Checks if the calling task has been notified of a pending reset
**
** PARAMETERS: isResource -
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PRIVATE BOOL appOsIsResetPendingCheck( BOOL isResource )
{
   AOS_TASKSTATE  state;
   OS_STATUS      status;

   status = appOsGetTaskStateSelf( &state );
   if ( status != OS_STATUS_OK )
   {
      return ( FALSE );
   }

   /* Check state against current 'wake-up' condition states */
   if ( state == AOS_TASKSTATE_RESET_NTFY )
   {
      /* Interrupt the task because of a reset condition */
      return ( TRUE );
   }
   else if ( isResource && ( state == AOS_TASKSTATE_UNINTERRUPTIBLE ) )
   {
      /* Uninterruptible Tasks should not use interruptible resources */
      LOG(( "APPOS - ERROR: Uninterruptible task trying to use an interruptible resource!" ));
      ASSERT( 0 );
      /* return ( FALSE ); */
   }
   else
   {
      /* Task does not need to be interrupted */
      return ( FALSE );
   }
}

/*****************************************************************************
** FUNCTION:   appOsTaskResetNotifyAll
**
** PURPOSE:    This is a blocking call that informs all tasks of an application
**             reset and waits for each of them to ACK the reset.
**
** PARAMETERS: none
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PRIVATE OS_STATUS appOsTaskResetNotifyAll( void )
{
   OS_STATUS retval;
   AOSZTCB *t;
   int i;
   int nResetTasks = 0;

   LOG(( "APPOS: Enter appOsTaskResetNotifyAll " ));

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   /*
   ** Notify all tasks that are eligible of the pending reset
   */
   for ( i = 0; i < MAX_TCB; i++ )
   {
      t = &aoszcb.tcb[i];

      if ( t->inuse )
      {
         switch ( t->state )
         {
            case AOS_TASKSTATE_NORMAL:
            {
               appOsSetTaskState( t->taskid, AOS_TASKSTATE_RESET_NTFY );
               LOG(( "APPOS: appOsTaskResetNotifyAll informed task %s", appOsGetTaskName( t->taskid ) ));
               nResetTasks++;
            }
            break;

            case AOS_TASKSTATE_SUSPENDED:
            {
               /* Do not resume un-interruptible tasks, or notify them of the reset.
                * It is the parent task's responsibility to deal with un-interruptible
                * tasks during an application reset. */
               if ( t->savedState != AOS_TASKSTATE_UNINTERRUPTIBLE )
               {
                  LOG(( "APPOS: appOsTaskResetNotifyAll informed and resumed task %s", appOsGetTaskName( t->taskid ) ));

                  /* Directly resume the task, and set the task state. Calling
                   * appOsTaskResume would result in deadlock, since we have already
                   * taken the same mutex used by appOsTaskResume */
                  t->state = AOS_TASKSTATE_RESET_NTFY;
                  bcmOsEventSend( &t->events[ AOS_RESUME_EVENT ] );
                  nResetTasks++;
               }
            }
            break;

            case AOS_TASKSTATE_UNINTERRUPTIBLE:
            default:
            {
               /* Do not interrupt */
            }
            break;
         }
      }
   }

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   /*
   ** Wait for all tasks notified of reset to acknowledge
   */
   LOG(( "APPOS: %d tasks notified of reset, waiting for acks...\n", nResetTasks ));
   for ( i = 0; i < nResetTasks; i++ )
   {
      retval = bcmOsSemTake( aoszcb.taskResetSemId, OS_WAIT_FOREVER );

      if ( retval != OS_STATUS_OK )
      {
         LOG(( "appOsTaskResetNotifyAll(): BIG TROUBLE in little china! Can't take a semaphore while waiting for reset ack!" ));
      }
   }

   LOG(( "\nAPPOS: All %d tasks acknowledged reset. ", nResetTasks ));

   LOG(( "APPOS: Exit appOsTaskResetNotifyAll " ));
   return ( OS_STATUS_OK );
}


/*****************************************************************************
** FUNCTION:   appOsTaskResetNotify
**
** PURPOSE:    This is a blocking call that informs a task of a reset
**             and waits for the task to ACK the reset.
**
** PARAMETERS: none
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PUBLIC OS_STATUS appOsTaskResetNotify( OS_TASKID taskid )
{
   OS_STATUS retval;
   AOSZTCB *t;

   LOG(( "APPOS: Enter appOsTaskResetNotify " ));

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   /*
   ** Notify all tasks that are eligible of the pending reset
   */
   retval = appOsGetTCB( taskid, &t );
   if ( retval != OS_STATUS_OK )
   {
      /* Leave critical section */
      bcmOsMutexUnlock( aoszcb.tcbMutexId );
      return ( retval );
   }

   if ( t->inuse )
   {
      switch ( t->state )
      {
         case AOS_TASKSTATE_NORMAL:
         {
            appOsSetTaskState( t->taskid, AOS_TASKSTATE_RESET_NTFY );
            LOG(( "APPOS: appOsTaskResetNotify informed task %s", appOsGetTaskName( t->taskid ) ));
         }
         break;

         case AOS_TASKSTATE_SUSPENDED:
         {
            /* Do not resume un-interruptible tasks, or notify them of the reset.
             * It is the parent task's responsibility to deal with un-interruptible
             * tasks during an application reset. */
            if ( t->savedState != AOS_TASKSTATE_UNINTERRUPTIBLE )
            {
               LOG(( "APPOS: appOsTaskResetNotify informed and resumed task %s", appOsGetTaskName( t->taskid ) ));

               /* Directly resume the task, and set the task state. Calling
                * appOsTaskResume would result in deadlock, since we have already
                * taken the same mutex used by appOsTaskResume */
               t->state = AOS_TASKSTATE_RESET_NTFY;
               bcmOsEventSend( &t->events[ AOS_RESUME_EVENT ] );
            }
         }
         break;

         case AOS_TASKSTATE_UNINTERRUPTIBLE:
         {
            /* Do not interrupt */
            LOG(( "APPOS: appOsTaskResetNotify can't reset an uninterruptible task!" ));

            /* Leave critical section */
            bcmOsMutexUnlock( aoszcb.tcbMutexId );

            return ( OS_STATUS_ERR );
         }
         break;

         default:
         {
            /* Leave critical section */
            bcmOsMutexUnlock( aoszcb.tcbMutexId );

            return ( OS_STATUS_OK );
         }
         break;
      }
   }

   /* Set this task as the source of the reset */
   retval = bcmOsTaskIdentSelf( &t->resetSrcTaskid );
   if ( retval != OS_STATUS_OK )
   {
      /* Leave critical section */
      bcmOsMutexUnlock( aoszcb.tcbMutexId );
      return ( retval );
   }

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   /*
   ** Wait for the task notified of reset to acknowledge
   */
   LOG(( "APPOS: Task 0x%x notified of reset, waiting for ack... ", taskid ));
   retval = appOsWaitForEvent( AOS_RESET_ACK_EVENT );

   if ( retval != OS_STATUS_OK )
   {
      LOG(( "appOsTaskResetNotify(): BIG TROUBLE in little china! Can't wait on event while waiting for reset ack!" ));
   }
   else
   {
      LOG(( "APPOS: Task 0x%x acknowledged reset. ", taskid ));
   }

   LOG(( "APPOS: Exit appOsTaskResetNotify " ));
   return ( OS_STATUS_OK );
}


/*****************************************************************************
** FUNCTION:   appOsTaskResetAck
**
** PURPOSE:    This function should be invoked by tasks to acknowledge that
**             they have been notified that a reset is pending. Calling this
**             function will cause the calling task to block until it is
**             placed into the shutdown state.
**
** PARAMETERS: none
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PUBLIC OS_STATUS appOsTaskResetAck( void )
{
   OS_STATUS retval;
   OS_TASKID taskid;
   AOSZTCB *t;

   LOG(( "APPOS: Enter appOsTaskResetAck (task=%s) ", appOsGetTaskNameSelf() ));

   retval = bcmOsTaskIdentSelf( &taskid );
   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }
   retval = appOsGetTCB( taskid, &t );
   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   retval = appOsSetTaskStateSelf( AOS_TASKSTATE_RESET_ACKED );

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   /*
   ** Notify the reset source that we have acknowledged the reset condition
   */
   if ( t->resetSrcTaskid )
   {
      appOsSignalEvent( AOS_RESET_ACK_EVENT, t->resetSrcTaskid );
      t->resetSrcTaskid = 0;
   }
   else
   {
      /* Notify the mta task that we have acked the reset_ntfy */
      retval = bcmOsSemGive( aoszcb.taskResetSemId );
   }

   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }


   /* Wait until told to resume by the mta root task */
   retval = appOsWaitForEvent( AOS_TASKSHUTDOWN_EVENT );

   LOG(( "APPOS: Exit appOsTaskResetAck (task=%s) ", appOsGetTaskNameSelf() ));

   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   retval = appOsSetTaskStateSelf( AOS_TASKSTATE_SHUTDOWN );

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   return ( retval );
}


/*****************************************************************************
** FUNCTION:   appOsTaskProceedToShutdown
**
** PURPOSE:    This function will change the state of a task to shutdown
**             (from reset-ack). Tasks are blocked in the reset-ack state.
**             Changing the task's state to shutdown will cause the task to
**             resume operation. In this state, tasks should perform operations
**             required for task shutdown before exiting their main task function.
**
** PARAMETERS: taskid - change this task to shutdown
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PUBLIC OS_STATUS appOsTaskProceedToShutdown( OS_TASKID taskid )
{
   OS_STATUS retval;
   AOS_TASKSTATE state;
   BOOL bSignalTask;

   LOG(( "APPOS: Enter appOsTaskProceedToShutdown 0x%x (%s)", taskid, appOsGetTaskName( taskid ) ));

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   retval = appOsGetTaskState( taskid, &state );
   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   switch ( state )
   {
      case AOS_TASKSTATE_RESET_ACKED:
      case AOS_TASKSTATE_RESET_NTFY:
      {
         /* Signal the task to go to shutdown */
         LOG(( "APPOS: appOsTaskProceedToShutdown() signaling task (%s)", appOsGetTaskName( taskid ) ));
         bSignalTask = TRUE;
      }
      break;

      case AOS_TASKSTATE_DONE:
      {
         /* Task already done, don't signal the task */
         LOG(( "APPOS: appOsTaskProceedToShutdown() not signaling task (%s) (task already DONE)", appOsGetTaskName( taskid ) ));
         bSignalTask = FALSE;
      }
      break;

      default:
      {
         /* Other states will do nothing: don't signal the task */
         LOG(( "APPOS: appOsTaskProceedToShutdown() not signaling task (%s) (not in RESET_ACKED or NTFY)", appOsGetTaskName( taskid ) ));
         bSignalTask = FALSE;
      }
      break;
   }

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   LOG(( "APPOS: Exit appOsTaskProceedToShutdown (%s)", appOsGetTaskName( taskid ) ));

   if ( bSignalTask )
   {
      /* Wake up the task to perform its shutdown operations */
      return ( appOsSignalEvent( AOS_TASKSHUTDOWN_EVENT, taskid ));
   }
   else
   {
      return ( OS_STATUS_OK );
   }
}


/*****************************************************************************
** FUNCTION:   appOsTaskEnterShutdown
**
** PURPOSE:    This is called invoked when a taskMain has completed before
**             entering the taskDeinit.
**
** PARAMETERS: none
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PRIVATE OS_STATUS appOsTaskEnterShutdown( void )
{
   OS_STATUS retval;
   OS_TASKID taskid;
   AOS_TASKSTATE state;

   LOG(( "APPOS: Enter appOsTaskEnterShutdown (task=%s) ", appOsGetTaskNameSelf() ));

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   retval = bcmOsTaskIdentSelf( &taskid );
   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   appOsGetTaskState( taskid, &state );

   if ( state == AOS_TASKSTATE_NORMAL )
   {
      /* A running task has exited its main. Change its state to shutdown */
      appOsSetTaskState( taskid, AOS_TASKSTATE_SHUTDOWN );
   }
   else if ( state == AOS_TASKSTATE_RESET_NTFY )
   {
      /* The task was notified of a reset, but didn't ack it.
      ** We must ack the reset, since the notifier is waiting for this ack.
      ** We must assume it was in the process of exiting on its own so we will
      ** allow it to proceed with its shutdown. This is accomplished by sending
      ** the shutdown event before acking, so that reset ack will not block. */
      appOsTaskProceedToShutdown( taskid );
      appOsTaskResetAck();
   }

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   LOG(( "APPOS: Exit appOsTaskEnterShutdown (task=%s) ", appOsGetTaskNameSelf() ));
   return ( OS_STATUS_OK );
}


/*****************************************************************************
** FUNCTION:   appOsTaskEnterDone
**
** PURPOSE:    This is a blocking call invoked when a task has completed
**             execution. The calling task will wait until it is re-spawned.
**
** PARAMETERS: none
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PRIVATE OS_STATUS appOsTaskEnterDone( void )
{
   OS_STATUS retval;

   LOG(( "APPOS: Enter appOsTaskEnterDone (task=%s) ", appOsGetTaskNameSelf() ));

   /* Free resources that are "owned" by this task, but it is not aware of */
   retval = appOsFreeSecretResources();
   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   /* Change own task's state to reset_done */
   retval = appOsSetTaskStateSelf( AOS_TASKSTATE_DONE );
   if ( retval != OS_STATUS_OK )
   {
      return ( retval );
   }

   LOG(( "APPOS: Exit appOsTaskEnterDone (task=%s) ", appOsGetTaskNameSelf() ));

   /* Wait until told to resume by the mta root task */
   return ( appOsWaitForEvent( AOS_RESURRECT_EVENT ));
}


/*****************************************************************************
** FUNCTION:   appOsFreeSecretResources
**
** PURPOSE:    This function will free resources that are owned but were not
**             allocated by the calling task. i.e. the task is unaware that
**             it owns the resources and therefore did not de-allocate them
**             in it's task de-initialization function.
**
** PARAMETERS: none
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PRIVATE OS_STATUS appOsFreeSecretResources( void )
{
   /* It is possible for another task to create a socket, and then share that
    * socket with this task. This can be done without the knowledge of this
    * task. In this case, this task is considered to have co-ownership of the
    * socket and must de-allocate it. Since this task did not directly allocate
    * the socket, and has no knowledge of it, we need to free the sockets here. */
   return ( bcmOsSocketCloseShared() );
}


/*****************************************************************************
** FUNCTION:   appOsTaskWaitForDone
**
** PURPOSE:    This function will block until the specified task is "done".
**             A task is considered done once it has exited from it's main
**             task function and completed execution of the task de-init
**             function.
**
** PARAMETERS: taskid - wait until this task is done
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PUBLIC OS_STATUS appOsTaskWaitForDone( OS_TASKID taskid )
{
   AOS_TASKSTATE state;

   LOG(( "APPOS: Enter appOsTaskWaitForDone - %s waiting on task 0x%x (%s)",
         appOsGetTaskNameSelf(), taskid, appOsGetTaskName( taskid ) ));

   while ( 1 )
   {
      /* Enter critical section */
      bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

      appOsGetTaskState( taskid, &state );

      /* Leave critical section */
      bcmOsMutexUnlock( aoszcb.tcbMutexId );

      if ( state == AOS_TASKSTATE_DONE )
      {
         break;
      }
      else
      {
         bcmOsSleep( ONE_HUNDRED_MSEC );
      }
   }

   LOG(( "APPOS: Exit appOsTaskWaitForDone - %s done waiting on task 0x%x (%s)",
         appOsGetTaskNameSelf(), taskid, appOsGetTaskName( taskid ) ));

   return ( OS_STATUS_OK );
}


/*****************************************************************************
** FUNCTION:   appOsTaskResurrect
**
** PURPOSE:    To re-spawn a task that was previously reset
**
** PARAMETERS: tcb - Task control block of the task to resurrect
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       It is the caller's responsibility to insure concurrency protection
*****************************************************************************/
PRIVATE OS_STATUS appOsTaskResurrect( AOSZTCB *tcb )
{
   /* Change task's state to normal */
   tcb->state = AOS_TASKSTATE_NORMAL;

   /* Resume the previously suspended task */
   return ( bcmOsEventSend( &tcb->events[ AOS_RESURRECT_EVENT ] ));
}


/*****************************************************************************
** FUNCTION:   appOsStartApp
**
** PURPOSE:    This function is responsible for spawning the application. The
**             application initialization function will be invoked asynchronously.
**             This function should spawn all necessary application tasks,
**             allocate resources, and init local storage. The application
**             de-initialization function will be invoked when an application
**             reset is initiated. This function should wait for tasks to
**             complete the reset, and de-allocate resources.
**
** PARAMETERS: appInit - application initialization function
**             appDeinit - application deinitialization function
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PUBLIC void appOsStartApp( AOS_APP_INIT_FNC appInit, AOS_APP_DEINIT_FNC appDeinit )
{
   LOG(( "APPOS: Enter appOsStartApp" ));

   aoszcb.appTask.init = appInit;
   aoszcb.appTask.deinit = appDeinit;
   aoszcb.appTask.stackSize = (8*1024);
   aoszcb.appTask.taskClass = APPOS_CFG_APP_TASK_PRIORITY;
   aoszcb.appTask.taskArg = 0;
   aoszcb.appTask.name = AOS_APPTASK_NAME;

   bcmOsTaskCreate( AOS_ROOTTASK_NAME,
                    (4*1024),
                    APPOS_CFG_ROOT_TASK_PRIORITY,
                    appOsRootTask,
                    0,
                    &aoszcb.rootTaskId );

   LOG(( "APPOS: Exit appOsStartApp" ));

}


/*****************************************************************************
** FUNCTION:   appOsRootTask
**
** PURPOSE:    Main function for the root task. The root task spawns the
**             application task, waits for an application reset, informs
**             all tasks of the reset, waits for an ACK from each task, resumes
**             the application task, and then waits for the reset to complete,
**             ad-infinitum.
**
** PARAMETERS: taskArg - not used
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PRIVATE void appOsRootTask( OS_TASKARG taskArg )
{
   while ( 1 )
   {
      LOG(( "\n\nAPPOS: appOsRootTask() - Is it morning already? Spawning app task (epoch #%d)...\n", aoszcb.nAppResets ));

      /* Spawn the application task */
      appOsTaskCreate(  aoszcb.appTask.name,
                        aoszcb.appTask.stackSize,
                        aoszcb.appTask.taskClass,
                        aoszcb.appTask.init,
                        appOsAppTaskMain,
                        aoszcb.appTask.deinit,
                        aoszcb.appTask.taskArg,
                        &aoszcb.appTask.taskId );

      /*
      ** Wait forever for an application reset
      ** NOTE: it IS valid for a reset to occur during the appTask.init, this
      ** is why we must initialize the app asynchronously and wait for the reset
      ** signal immediately
      */

      /* This be done in the appOs, but until we resolve the clear dhcp IP
       * problem, ccInit() will enable reset detection before provisioning */
#if 0
      /* Reset detection should be enabled by default */
      appOsResetDetectionEnable();
#endif

      /* Wait forever for an application reset */
      bcmOsSemTake( aoszcb.appResetSemId, OS_WAIT_FOREVER );
      aoszcb.nAppResets++;
      LOG(( "\n\nAPPOS: appOsRootTask() - RESET NOW (reset #%d)\n", aoszcb.nAppResets ));

      /* Notify all tasks of the reset, blocking until all have acked */
      appOsTaskResetNotifyAll();

      /* De-init the application */
      appOsTaskProceedToShutdown( aoszcb.appTask.taskId );

      /* Wait for the reset to complete */
      appOsTaskWaitForDone( aoszcb.appTask.taskId );

      /* Application reset is complete */
      aoszcb.resetState = AOS_RESETSTATE_NORESET;

      LOG(( "\nAPPOS: appOsRootTask() - Reset #%d completed, good night...\n\n", aoszcb.nAppResets ));

      /* Repeat ad-infinitum */
   }
}


/*****************************************************************************
** FUNCTION:   appOsAppTaskMain
**
** PURPOSE:    This the main task function for the application task. It
**             will be invoked after the application initialization function
**             is invoked. This function will suspend the application task
**             until it is resumed (by the root task). The app task is only
**             resumed after an application reset has been initiated. After
**             being resumed, the appliation task will ACK the reset, and
**             return. This will cause the application de-initialization
**             function to be invoked.
**
** PARAMETERS: taskArg - non used
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
*****************************************************************************/
PRIVATE void appOsAppTaskMain( OS_TASKARG taskArg )
{
   OS_STATUS      retval;
   AOS_TASKSTATE  state;

   /* Suspend the app task, we will be woken up by a reset.
   ** (If a reset has already been notified, this will return immediately) */
   retval = appOsTaskSuspend( 0 );

   if ( retval == AOS_STATUS_RESET )
   {
      /* Acknowledge the reset, waiting until told to proceed */
      appOsTaskResetAck();
   }
   else
   {
      /* A reset is the only thing that should resume this task */
      LOG(( "ERROR: appOsAppTaskMain() has been resumed illegally!" ));
      ASSERT( 0 );
   }

   appOsGetTaskStateSelf( &state );
   if ( state == AOS_TASKSTATE_SHUTDOWN )
   {
      LOG(( "App task has entered the shutdown state, exiting main..." ));
   }
   else
   {
      /* Should never be here */
      LOG(( "ERROR: App task resumed, but not in the shutdown state, exiting main..." ));
   }
}


/*****************************************************************************
** FUNCTION:   appOsResetApp
**
** PURPOSE:    Initiate an application reset
**
** PARAMETERS: none
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       The application reset is performed asynchronously, this function
**             returns immediately
*****************************************************************************/
PUBLIC void appOsResetApp( void )
{
   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   switch ( aoszcb.resetDetectMode )
   {
      case AOS_RESETDETECT_ENABLED:
      {
         LOG(( "APPOS: appOsResetApp() - Signalling root task of reset." ));

         /* Reset detection is enabled, disable detection to avoid a duplicate
         ** reset signal, and then notify the aoRootTask of the reset */
         aoszcb.resetDetectMode = AOS_RESETDETECT_DISABLED;
         aoszcb.resetState = AOS_RESETSTATE_RESETTING;
   bcmOsSemGive( aoszcb.appResetSemId );
}
      break;

      case AOS_RESETDETECT_DISABLED:
      {
         /* Ignore the reset if detection is disabled */
         LOG(( "APPOS: appOsResetApp() - Ignoring reset, reset detection is disabled." ));
      }
      break;

      case AOS_RESETDETECT_INHIBITED:
      {
         /* Postpone handling the reset if detection is inhibited */
         LOG(( "APPOS: appOsResetApp() - Postponing reset until detection is enabled, reset detection is currently inhibited." ));
         aoszcb.resetState = AOS_RESETSTATE_POSTPONED;
      }
      break;
   }

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );
}


/*****************************************************************************
** FUNCTION:   appOsResetDetectionEnable
**
** PURPOSE:    Enable reset detection
**
** PARAMETERS: none
**
** RETURNS:    none
**
** NOTE:
*****************************************************************************/
PUBLIC void appOsResetDetectionEnable( void )
{
   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   switch ( aoszcb.resetDetectMode )
   {
      case AOS_RESETDETECT_ENABLED:
      {
         LOG(( "APPOS: appOsResetDetectionEnable() - Doing nothing, detection already enabled." ));
      }
      break;

      case AOS_RESETDETECT_DISABLED:
      {
         LOG(( "APPOS: appOsResetDetectionEnable() - Enabled reset detection." ));
         aoszcb.resetDetectMode = AOS_RESETDETECT_ENABLED;
      }
      break;

      case AOS_RESETDETECT_INHIBITED:
      {
         aoszcb.resetDetectMode = AOS_RESETDETECT_ENABLED;

         if ( aoszcb.resetState != AOS_RESETSTATE_POSTPONED )
         {
            LOG(( "APPOS: appOsResetDetectionEnable() - Enabled reset detection, no resets were postponed." ));
         }
         else
         {
            /* A reset was postponed, signal it now */
            LOG(( "APPOS: appOsResetDetectionEnable() - Enabled reset detection, continuing a postponed reset..." ));
            aoszcb.resetState = AOS_RESETSTATE_NORESET;
            appOsResetApp();
         }
      }
      break;
   }

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );
}


/*****************************************************************************
** FUNCTION:   appOsResetDetectionDisable
**
** PURPOSE:    Disable reset detection (resets will be ignored)
**
** PARAMETERS: none
**
** RETURNS:    none
**
** NOTE:
*****************************************************************************/
PRIVATE void appOsResetDetectionDisable( void )
{
   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   switch ( aoszcb.resetDetectMode )
   {
      case AOS_RESETDETECT_DISABLED:
      {
         LOG(( "APPOS: appOsResetDetectionDisable() - Reset detection is already disabled." ));
      }
      break;

      case AOS_RESETDETECT_ENABLED:
      {
         LOG(( "APPOS: appOsResetDetectionDisable() - Reset detection has been disabled." ));
      }
      break;

      case AOS_RESETDETECT_INHIBITED:
      {
         LOG(( "APPOS: appOsResetDetectionDisable() - Reset detection has been disabled." ));
      }
      break;
   }
   aoszcb.resetDetectMode = AOS_RESETDETECT_DISABLED;

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );
}


/*****************************************************************************
** FUNCTION:   appOsResetDetectionInhibit
**
** PURPOSE:    Inhibit reset detection (resets won't be processed until enabled)
**
** PARAMETERS: none
**
** RETURNS:    OS_STATUS_OK      - if able to inhibit resets
**             AOS_STATUS_RESET  - if a reset is already in progress
**
** NOTE:
*****************************************************************************/
PUBLIC OS_STATUS appOsResetDetectionInhibit( void )
{
   OS_STATUS retval = OS_STATUS_OK;

   /* Enter critical section */
   bcmOsMutexLock( aoszcb.tcbMutexId, OS_WAIT_FOREVER );

   if ( aoszcb.resetState == AOS_RESETSTATE_RESETTING )
   {
      LOG(( "APPOS: appOsResetDetectionInhibit() - Can't inhibit reset detection, reset already in progress." ));
      retval = AOS_STATUS_RESET;
   }
   else
   {
      aoszcb.resetDetectMode = AOS_RESETDETECT_INHIBITED;
      LOG(( "APPOS: appOsResetDetectionInhibit() - Reset detection has been inhibited." ));
   }

   /* Leave critical section */
   bcmOsMutexUnlock( aoszcb.tcbMutexId );

   return ( retval );
}
