/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: net.c
*    Creation Date: 18 Aug 2000 (v0.00)
*    VSS Info:
*        $Revision: 1.2 $
*        $Date: 2002/12/23 20:31:35 $
*
****************************************************************************
*    Description:
*
*      Network utility functions
*
****************************************************************************/

/*
*****************************************************************************
** INCLUDE FILES
*****************************************************************************
*/

//#include <hausenv.h>
#include <net.h>
#include <string.h>
#include <stdlib.h>
///#include <bcmos.h>

#include <bcmosHelper.h>

/*
*****************************************************************************
** LOCAL FUNCTIONS
*****************************************************************************
*/
static IPADDR  IpAddrNumber( const char *string );


/*****************************************************************************
** FUNCTION:   bcmNetInit
**
** PURPOSE:    Initialize the NET module
**
** PARAMETERS: none
**
** RETURNS:    NET_STATUS_OK
**
** NOTE:       This function was added for consistency with the rest
**             of the module initialization code
*****************************************************************************
*/
#if CH_NEEDED
NET_STATUS bcmNetInit( void )
{
   return ( NET_STATUS_OK );
}
#endif

/*
*****************************************************************************
** FUNCTION:   bcmNetDeinit
**
** PURPOSE:    Shutdown the NET module
**
** PARAMETERS: none
**
** RETURNS:    NET_STATUS_OK
**
** NOTE:       This function was added for consistency with the rest
**             of the module de-initialization code
*****************************************************************************
*/
#if CH_NEEDED
NET_STATUS bcmNetDeinit( void )
{
   return ( NET_STATUS_OK );
}
#endif

/*
*****************************************************************************
** FUNCTION:   IpAddrNumber()
**
** PURPOSE:    Converts character represented IP address (dotted decimal notation)
**             to a four byte integer. e.g. converts "10.1.0.5" to 0x0a010005.
**
** PARAMETERS: string - Character represented IP adress (dottted decimal notion).
**
** RETURNS:    "IPADDR" IP address or 0 if error
**
** NOTE:
*****************************************************************************
*/
static IPADDR  IpAddrNumber( const char *string )
{
   char     ipstring[256];  /* It might be FQDN with the MAX length = 256 (as per spec) */
   char     *s = ipstring;
   IPADDR   address = 0L;
   int      i, j;
   char     *cp, *b1, *b2, *cl;

   if (!string)
   {
      LOG(("*PANIC* IpAddrNumber() - NULL String as an input parameter!\n"));
      return 0;
   }
   if (!strlen(string))
   {
      LOG(("*PANIC* IpAddrNumber() - EMPTY String as an input parameter!\n"));
      return 0;
   }

   strcpy(ipstring, string);  /* Just make the string possible to manipulate/change with. */

   /*
   ** According to the PacketCable conventions, explicit IP addresses
   ** are presented in the squuare brackets - [xx.xx.xx.xx]
   ** Let's get rid of these (if they are present)
   ** Another convention is that the FQDN can contain port number in it,
   ** like this: <FQDN>:<port_num>. We need to get rid of it as well.
   */
   b1 = strchr( ipstring, '[');
   b2 = strchr( ipstring, ']');
   cl = strchr( ipstring, ':');

   if ( b1 )
   {
      b1++;    /* skip to the next character */
   }

   if (b1  && (b1 >= b2) )
   {
      LOG(("*PANIC* IpAddrNumber() - WRONG Format of an input parameter (%s)!\n", string));
      return 0;
   }

   if (b2  && !b1 )
   {
      LOG(("*PANIC* IpAddrNumber() - WRONG Format of an input parameter (%s)!\n", string));
      return 0;
   }

   if ( cl && (cl < b2) )
   {
      LOG(("*PANIC* IpAddrNumber() - WRONG Format of an input parameter (%s)!\n", string));
      return 0;
   }

   if (cl)
   {
      *cl = 0; /* Cut off everything starting from ':' sign. */
   }

   if (b2)     /* Cut off everything starting from ']' sign. */
   {
      *b2 = 0;
   }

   if (!b1)
   {
      b1 = s;
   }

   for (j=24; j>=0; j-=8)
   {
      cp = b1;

      while(( *b1 ) && ( *b1 != '.' ))
      {

         if ( (*b1 < '0') || (*b1 > '9') )
         {
            address = 0L;
            return(address);
         }

         b1++;
      }

      if( *b1 )
      {
         *b1++ = '\0';
      }

      i = atoi(cp);
      address |= (i &0xff) << j;

      if ( !(*b1) && (j != 0) )
      {
         LOG(("*PANIC* IpAddrNumber() - WRONG Format of an input parameter (%s)!\n", string));
         return 0;
      }

   }

   if ( *b1 )
   {
      LOG(("*PANIC* IpAddrNumber() - WRONG Format of an input parameter (%s)!\n", string));
      return 0;
   }

   return(address);
}

/*
*****************************************************************************
** FUNCTION:   bcmNetNumToIpAddr()
**
** PURPOSE:    Converts IPADDR type (four byte int) to a dotted decimal
**             notation IP address string. e.g. converts 0x0a010005 to
**             "10.1.0.5".
**
** PARAMETERS: ipaddr - IPADDR IP address
**             buf    - string containing IP Address (output)
**
** RETURNS:    NET_STATUS_OK or error code
**
** NOTE:
*****************************************************************************
*/
NET_STATUS bcmNetNumToIpAddr( IPADDR ipaddr, char *buf )
{
   sprintf(buf, "%d.%d.%d.%d", (ipaddr>>24)&0xff,
                               (ipaddr>>16)&0xff,
                               (ipaddr>>8)&0xff,
                               (ipaddr&0xff));

   return ( NET_STATUS_OK );
}


/*
*****************************************************************************
** FUNCTION:   bcmNetIpAddrToNum()
**
** PURPOSE:    Converts character represented IP address (dotted decimal notation)
**             or fully qualified domain name (FQDN) to an IPADDR type (four byte
**             integer).
**
** PARAMETERS: string - character string with IP address/FQDN
**             ipAddr - ip address (four byte integer)
**
** RETURNS:    NET_STATUS_OK or error
**
** NOTE:       This function will first try to convert the passed in string
**             assuming that it is a dotted decimal notion IP address. If
**             this conversion fails, it will perform a DNS lookup. If this
**             fails, it will then return an error.
** 
**             It is assumed that 'string' only contains the FQDN, and does not
**             contain any prefixed usernames. e.g. this function will not resolve
**             a string of the format "who-ever@where-ever.com", it will only 
**             resolve strings of the format "where-ever.com". It is responsibility
**             of the caller to strip prefixed usernames.
**
**             Use of this function is deprecated, and is only provided for
**             backwards compatibility. bcmNetGetHostAddrs should be used instead.
**
*****************************************************************************
*/
NET_STATUS  bcmNetIpAddrToNum( const char *string, IPADDR *ipAddr )
{
   OS_STATUS   status;

   /* Validate input parameter */
   if ( string == NULL )
   {
      return ( NET_STATUS_ERROR );
   }

   /* Check if domain name is in IP dotted notation, e.g. [10.1.100.1] */
   *ipAddr = IpAddrNumber( string );
   if ( *ipAddr == 0 )
   {
      printf(" gethostaddr removed for ch\n");
      return ( NET_STATUS_ERROR );
      #if CH_NEEDED
      /* Domain name not in IP dotted notation, perform DNS lookup */
      status = bcmOsDnsGetHostAddr((char *) string, ipAddr );
      if ( status != OS_STATUS_OK )
      {
         *ipAddr = 0;
         LOG(( "ERROR resolving address ""%s"" by DNS Server.", string ));
         return ( NET_STATUS_ERROR );
      }
      #endif
   }

   return( NET_STATUS_OK );
}


/*
****************************************************************************
** FUNCTION:   bcmNetGetHostAddrs
**
** PURPOSE:    Given the hostname, this function finds corresponding IP
**             addresses and returns them.
**
** PARAMETERS: hostName [IN]        - host name to resolve into IP addresses
**             ipAddr   [OUT]       - array of resolved IP addresses (in
**                                     host byte order). Memory must be
**                                     allocated by caller.
**             numIpAddrs [IN/OUT]  - On input, points to the array length
**                                     of 'ipAddr'. On output, returns the
**                                     number of resolved IP addresses contained
**                                     in the 'ipAddr' array.
**             useCache [IN]        - indicates whether to use the DNS cache
**
** RETURNS:    NET_STATUS_OK or error code
**
** NOTE:       The 'hostName' may be either a fully qualified domain name or
**             a dotted decimal notation IP address string.
** 
**             It is assumed that 'hostName' only contains the FQDN, and does not
**             contain any prefixed usernames. e.g. this function will not resolve
**             a string of the format "who-ever@where-ever.com", it will only 
**             resolve strings of the format "where-ever.com". It is responsibility
**             of the caller to strip prefixed usernames.
**
*****************************************************************************
*/
#if CH_NEEDED
NET_STATUS bcmNetGetHostAddrs( char *hostName, IPADDR *ipAddr, int *numIpAddrs, BOOL useCache )
{
   IPADDR      address;
   NET_STATUS  status = NET_STATUS_OK;

   /* Validate input parameters */
   ASSERT( hostName && ipAddr && numIpAddrs );

   /* Check if domain name is in IP dotted notation, e.g. [10.1.100.1] */
   address = IpAddrNumber( hostName );
   if ( address )
   {
      ASSERT( *numIpAddrs >= 1);

      /* Domain name is in IP dotted notation */
      ipAddr[0] = address;
      *numIpAddrs = 1;
   }
   else
   {
      printf(" gethostaddrs removed for ch\n");
      #if CH_NEEDED
      /* Domain name not in IP dotted notation, perform DNS lookup */
      if ( bcmOsDnsGetHostAddrs( hostName, ipAddr, numIpAddrs, useCache ) != OS_STATUS_OK )
      {
         *numIpAddrs = 0;
         status = NET_STATUS_ERROR;
         LOG(( "ERROR resolving address ""%s"" by DNS Server.", hostName ));
      }
      #endif
   }

   return( status );
}
#endif

/*
*****************************************************************************
** FUNCTION:   bcmNetGetOwnIpAddress()
**
** PURPOSE:    Retrieves the IP address of a given IP stack number
**
** PARAMETERS: netIface - IP Interface Number
**             ipAddr - IP address of interface number
**
** RETURNS:    NET_STATUS_OK or error code
**
** NOTE:       Parameter always defaults to Iface# = 1 if set to 0.
*****************************************************************************
*/
#if CH_NEEDED
NET_STATUS bcmNetGetOwnIpAddress( int netIface, IPADDR *ipAddr )
{
   *ipAddr = 0;
   if ( bcmOsGetOwnIpAddress( netIface, ipAddr ) != OS_STATUS_OK )
   {
      return ( NET_STATUS_ERROR );
   }

   return ( NET_STATUS_OK );
}
#endif

/*
*****************************************************************************
** FUNCTION:   bcmNetGetPortNum
**
** PURPOSE:    Extracts the Port Number part from the IP-Address string.
**             IP-Address string should be like <IP-str>:<port_num>
**
** PARAMETERS: ipAddrStr [IN]  - IP Addr string (or FQDN)
**             port      [OUT] - port number
**
** RETURNS:    NET_STATUS_OK or error code
**
** NOTE:
*****************************************************************************
*/
#if CH_NEEDED
NET_STATUS bcmNetGetPortNum( const char* ipAddrStr, UINT32 *port )
{
   char  *p;

   /* Sanity check */
   if ( !ipAddrStr )
   {
      *port = 0;
      return ( NET_STATUS_ERROR );
   }

   p = strchr( ipAddrStr, ':' );
   if ( p )
   {
      *port = atol( ++p );
      return ( NET_STATUS_OK );
   }

   return ( NET_STATUS_ERROR );
}
#endif
