/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: msgq.c
*    Creation Date: 28 June 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:54:39 $
*
****************************************************************************
*    Description:
*
*     Broadcom Generic Operating System Functions. This file contains a VxWorks
*     specific implementation of message queues.
*
****************************************************************************/

#include <hausenv.h>
#include <string.h>
#include <bcmos.h>

/* Operating system specific includes */
#include <vxWorks.h>
#include <msgQLib.h>
#include "vxwrkdef.h"


/*
*****************************************************************************
** TYPE DEFINITIONS and CONSTANTS
*****************************************************************************
*/

/* Definitions for task options */
#define OS_Q_NO_WAIT             NO_WAIT


/*****************************************************************************
** FUNCTION:   bcmOsQueueSend
**
** PURPOSE:    Sends a message to a specified ordinary Message Queue.
**
** PARAMETERS: qid      - Message Queue ID to send a message to.
**             qMsg     - Message Buffer
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       bcmOsQueueSend does not block on send.
******************************************************************************/
OS_STATUS bcmOsQueueSend(OS_QID qid, OS_QMSG *qMsg)
{
   NATIVE_STATUS  status;

   /* PSOS limits the size of the messages within a queue to 16 bytes msgBuf[4], VXWORKS
    * has no limit.  The generic bcmOsQueueSend will limit the VXWORKS message sizes
    * to be 16 bytes so that the interface is consistent across different operating systems. */

   ASSERT( qMsg );

   status = msgQSend( (MSG_Q_ID) qid, (char *)qMsg->buffer, sizeof( OS_QMSG ), OS_Q_NO_WAIT, MSG_PRI_NORMAL );

   /* Error checking */
   return( checkStatus( status, "bcmOsQueueSend" ));
}


/*****************************************************************************
** FUNCTION:   bcmOsQueueReceive
**
** PURPOSE:    receives a messgae from a specified ordinary Message Queue
**
** PARAMETERS: qid      - Message Queue ID to send a message to.
**             timeout  - Timeout in clock ticks
**             qMsg     - Message Buffer
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE: If timeout is 0, bcmOsQueueReceive() will be non-blocking, i.e. it
**       will return even if there is no available message in the queue. If
**       timeout is not 0, then this function will block until a message is
**       posted to the queue or the timeout occurs, which ever comes first.
**
******************************************************************************/
OS_STATUS bcmOsQueueReceive(OS_QID qid, OS_TICKS timeout, OS_QMSG *qMsg)
{
   NATIVE_STATUS  status;
   int            tickWait;

   ASSERT( qMsg );

   /* Convert to OS-specific tick value */
   if ( timeout == OS_WAIT_FOREVER )
   {
      /* Wait forever */
      tickWait = WAIT_FOREVER;
   }
   else if ( timeout )
   {
      /* Wait until either message is received, or timeout expires */
      tickWait = timeout / OS_TICK_SCALE_FACTOR;
   }
   else
   {
      /* Non-blocking semaphore take */
      tickWait = NO_WAIT;
   }


   /* PSOS limits the size of the messages within a queue to 16 bytes msgBuf[4], VXWORKS
    * has no limit.  The generic bcmOsQueueCreate will limit the VXWORKS message sizes
    * to be 16 bytes so that the interface is consistent across different operating systems. */
   status = msgQReceive( (MSG_Q_ID) qid, (char *)qMsg->buffer, sizeof( OS_QMSG ), tickWait );

   /* Error checking */
   return( checkStatus( status, "bcmOsQueueReceive" ));
}


/*****************************************************************************
** FUNCTION:   bcmOsQueueCreate
**
** PURPOSE:    creates an ordinary message queue of size (qSize).
**
** PARAMETERS: qName    - Message Queue Name.
**             qSize    - Message Queue Size (number of messages in the queue)
**             qType    - Message Queue type - fifo or priority
**
**             qid      - Message Queue ID (returned)
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       PSOS limits the size of the messages within a queue to 16 bytes
**             msgBuf[4], VXWORKS has no limit.  The generic bcmOsQueueCreate
**             will limit the VXWORKS message size to be 16 bytes so that the
**             interface is consistent across different operating systems.
******************************************************************************/
OS_STATUS bcmOsQueueCreate(char* qName, OS_QSIZE qSize, OS_QUEUE_TYPE qType, OS_QID *qid)
{
   NATIVE_STATUS  status;
   int qFlgs = 0;

   ASSERT( qName && qid );

   if ( qType == OS_QUEUE_FIFO )
   {
      qFlgs = MSG_Q_FIFO;
   }
   else if ( qType == OS_QUEUE_PRIORITY )
   {
      qFlgs = MSG_Q_PRIORITY;
   }

   /* PSOS limits the size of the messages within a queue to 16 bytes msgBuf[4], VXWORKS
    * has no limit.  The generic bcmOsQueueCreate will limit the VXWORKS message sizes
    * to be 16 bytes so that the interface is consistent across different operating systems. */
   *qid = (OS_QID) msgQCreate( (int) qSize, sizeof( OS_QMSG ), qFlgs );

   /* msgQCreate() returns NULL for an error. This is inconsistent with most other
    * VxWorks API calls which return ERROR for an error. If NULL is returned,
    * change the return code to ERROR so that the same error checking function can
    * be used for all API calls. */
   status = OK;
   if ( *qid == 0 )
   {
      status = ERROR;
   }

   /* Error checking */
   return( checkStatus( status, "bcmOsQueueCreate" ));
}


/*****************************************************************************
** FUNCTION:   bcmOsQueueDelete
**
** PURPOSE:    Delete a message queue
**
** PARAMETERS: qid      - Message Queue ID
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
**
******************************************************************************/
OS_STATUS bcmOsQueueDelete( OS_QID qid )
{
   NATIVE_STATUS  status;
   status = msgQDelete( qid );

   /* Error checking */
   return( checkStatus( status, "bcmOsQueueDelete" ));
}
