/***************************************************************************
 *     Copyright 2001  Broadcom Corporation
 *     All Rights Reserved
 *     No portions of this material may be reproduced in any form without the
 *     written permission of:
 *              Broadcom Corporation
 *              16251 Laguna Canyon Road
 *              Irvine, California  92618
 *     All information contained in this document is Broadcom Corporation
 *     company private, proprietary, and trade secret.
 *
 ****************************************************************************
 *
 *     Filename:  SecNVRAM.c
 *     Creation Date: 20 June 2001 (v0.00)
 *     VSS Info:
 *         $Revision: 1.1 $
 *         $Date: 2002/08/21 17:53:48 $
 *
 ****************************************************************************
 *     Description:
 *
 *       Implementation of wrappers to inerface to NVRAM for security
 *
 *     Notes:
 *
 ****************************************************************************/

//#ifndef WIN32
//#include <cmmtaint.h>
//#endif
#include <stdlib.h> 
#include <ctype.h> 
#include "SecLib.h"
#include "hmac.h"

#ifdef DEMO
/* structs using in table to map info to NVRAM */
typedef struct
{
   unsigned char   t_hash[MD5_SIZE];          /* hash of realm */
   unsigned char   flag;                /* prov or cms */
   int             skew;                /* time skew value */
} nvram_skew_info;
#endif


typedef struct 
{
   unsigned char   t_hash[MD5_SIZE];          /* hash of realm */
   int             index;               
} nvram_kdc_index_info;

typedef struct 
{
   unsigned char   t_hash[MD5_SIZE];          /* hash of fqdn */
   int             index;                
} nvram_cms_index_info;

typedef struct
{
   int              cert_len;
   unsigned char    cert[MAX_CERTIFICATE_LENGTH];
} nvram_certificate;

typedef struct
{
   int              key_len;
   unsigned char    key[MAX_CERT_KEY_LENGTH];
} nvram_key;


/* fucntion prototypes */
static void secWriteNVRAMSkew(
      int             skew_value,
      int             flag,
      char            *realm,
      unsigned long   realm_length);
static int secReadNVRAMSkew(
      int             flag,
      char            *realm,
      unsigned long   realm_length);
static void secWriteNVRAMTkt(
      unsigned char   *tkt,
      int             tkt_len,
      int             flag,
      char            *fqdn,
      unsigned long   fqdn_length);
static unsigned long secReadNVRAMTkt(
      unsigned char   *tkt,
      unsigned long   tkt_len,
      int             flag,
      char            *fqdn,
      unsigned long   fqdn_length);
static void secWriteNVRAMSessionKeyAndTktTimes(
      unsigned char   *session_key,
      unsigned long   session_key_len,
      int             session_key_type,
      int             flag,
      char            *fqdn,
      unsigned long   fqdn_length,
      int             auth_time,
      int             start_time,
      int             end_time,
      int             renew_till);
static unsigned long secReadNVRAMSessionKeyAndTktTimes(
      unsigned char   *session_key,
      int             *session_key_type,
      int             flag,
      char            *fqdn,
      unsigned long   fqdn_length,
      int             *auth_time,
      int             *start_time,
      int             *end_time,
      int             *renew_till);
static unsigned long secReadNVRAMMTACertKeyData(
      SEC_CERT_TYPE   type,
      unsigned int    *len,
      unsigned char   *data);
static unsigned long secWriteNVRAMMTACertKeyData(
      SEC_CERT_TYPE   type,
      int             len,
      const unsigned char   *data);



/* assign functions to pointers */
static SEC_NVRAM_IFACE   secNVRAMIface =
{
   secWriteNVRAMSkew,
   secReadNVRAMSkew,
   secWriteNVRAMTkt,
   secReadNVRAMTkt,
   secWriteNVRAMSessionKeyAndTktTimes,
   secReadNVRAMSessionKeyAndTktTimes,
   secReadNVRAMMTACertKeyData,
   secWriteNVRAMMTACertKeyData,
};



/* tables to map information to VNRAM */
/* for time skew, max entries is 2 for provisioning (AS or AP) (2)
 * and 1 per endpoint per AS or AP for cms enrty (4*2) totals 10
 * for tkk, session key we need half of that.
 * Tkt, session key, 1 for prov and maximum of 4 per endpoints
 * for CMS totals 5
 * */
#define MAX_NVRAM_TABLE_ENTRIES         8


static   unsigned char tmp_tkt[MAX_TKT_LENGTH];
static   unsigned char tmp_skey[MAX_SESSION_KEY_LENGTH];
static   unsigned char tmp_fqdn[255];

#ifdef DEMO
/* certificates and keys */
static nvram_certificate         MTA_man_cert;
static nvram_certificate         MTA_dev_cert;
static nvram_key                 MTA_device_modulus_key;
static nvram_key                 MTA_device_priv_key;
static nvram_certificate         TEL_root_cert;

#define MAX_NVRAM_SKEW_TABLE_ENTRIES    10
static nvram_skew_info           skew_table[MAX_NVRAM_SKEW_TABLE_ENTRIES];
#endif



/*
*****************************************************************************
**
**  Function Name: SecLibGetNVRAMIface
**
**  PURPOSE:   return iface struct containing pointers to seclib functions
**
**  PARAMETERS:
**
**  RETURNS:
**
**  NOTES:
*****************************************************************************
*/
SEC_NVRAM_IFACE*   SecLibGetNVRAMIface(void)
{
   return &secNVRAMIface;
}


/*
******************************************************************************
** FUNCTION: find_kdc_index 
**
** PURPOSE:  Find an index for this realm 
**
** PARAMETERS:
**    1. realm
**    2. realm_length
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
static int find_kdc_index(unsigned char *realm, unsigned long realm_length)
{
#ifndef DEMO
   int              i;
   BOOL             rc;
   MTA_KdcEntry     kdc_entry;
   unsigned char    tmp_realm[255];

   kdc_entry.pRealmName = tmp_realm;
   for (i=0;i<MAX_NVRAM_TABLE_ENTRIES;i++)
   {
      rc = cmNvGetKdcEntry(i, &kdc_entry); 
      if ( rc )
      {
         if ( ! memcmp(kdc_entry.pRealmName, realm, realm_length))
         {
            return i;
         }
      }
   }
   LOG(("get_kdc_index: NO NVRAM KDC entries found for realm %s.", 
            realm));
   return -1;
#else
   return -1;
#endif
}


/*
******************************************************************************
** FUNCTION: get_kdc_index 
**
** PURPOSE:  Find an index for this realm or get a free index and return 
**
** PARAMETERS:
**    1. realm
**    2. realm_length
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
static int get_kdc_index(unsigned char *realm, unsigned long realm_length)
{
#ifndef DEMO
   int             ix;
   int             i;
   MTA_KdcEntry    kdc_entry;
   unsigned char   tmp_realm[255];
   BOOL            rc;

   kdc_entry.pRealmName = tmp_realm;
   ix = find_kdc_index(realm, realm_length);
   if ( ix < 0 )
   {
      LOG(("MTA Getting a free kdcEntry index."));
   }
   else
   {
      return ix;
   }

   /* find a free entry now */
   for (i=0;i<MAX_NVRAM_TABLE_ENTRIES;i++)
   {
      rc = cmNvGetKdcEntry(i, &kdc_entry); 
      if ( !rc )
      {
         LOG(("kdcEntry FREE index found: %d.", i));
         return i;
      }
   }
   if ( i == MAX_NVRAM_TABLE_ENTRIES )
   {
      LOG(("*PANIC* No free kdc entry available for realm = %s.", 
               realm));
      return -1;
   }
   return -1;
#else
   return -1;
#endif
}

/*
******************************************************************************
** FUNCTION: find_cms_index 
**
** PURPOSE:  Find an index for this fqdn
**
** PARAMETERS:
**    1. fqdn
**    2. fqdn_length
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
static int find_cms_index(unsigned char *fqdn, unsigned long fqdn_length)
{
#ifndef DEMO
   int              i;
   BOOL             rc;
   MTA_CmsEntry     cms_entry;

   cms_entry.pKerberosTicket = tmp_tkt;
   cms_entry.pKerberosSessionKey = tmp_skey;
   cms_entry.pCmsFqdn = tmp_fqdn;
   for (i=0;i<MAX_NVRAM_TABLE_ENTRIES;i++)
   {
      rc = cmNvGetCmsEntry(i, &cms_entry); 
      if ( rc )
      {
         if ( ! memcmp(cms_entry.pCmsFqdn, fqdn, fqdn_length))
         {
            return i;
         }
      }
   }
   LOG(("find_cms_index: NO NVRAM CMS entries found for fqdn %s.", fqdn));
   return -1;
#else
   return -1;
#endif
}

/*
******************************************************************************
** FUNCTION: get_cms_index 
**
** PURPOSE:  Find an index for this realm or get a free index and return 
**
** PARAMETERS:
**    1. fqdn
**    2. fqdn_length
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
static int get_cms_index(unsigned char *fqdn, unsigned long fqdn_length, int *new_entry)
{
#ifndef DEMO
   int             ix;
   int             i;
   MTA_CmsEntry    cms_entry;
   BOOL            rc;


   cms_entry.pKerberosTicket = tmp_tkt;
   cms_entry.pKerberosSessionKey = tmp_skey;
   cms_entry.pCmsFqdn = tmp_fqdn;
   ix = find_cms_index(fqdn, fqdn_length);
   if ( ix < 0 )
   {
      LOG(("MTA Getting a free CmsEntry index."));
      *new_entry = 1;
   }
   else
   {
      LOG(("get_cms_index index found %d.", ix));
      *new_entry = 0;
      return ix;
   }

   /* find a free entry now */
   for (i=0;i<MAX_NVRAM_TABLE_ENTRIES;i++)
   {
      rc = cmNvGetCmsEntry(i, &cms_entry);
      if ( !rc )
      /* if ( cms_entry.pCmsFqdn[0] == '\0' ) */
      {
         LOG(("cmsEntry FREE index found: %d.", i));
         return i;
      }
   }
   if ( i == MAX_NVRAM_TABLE_ENTRIES )
   {
      LOG(("*PANIC* No free NVRAM CMS entry available fqdn = %s.", fqdn));
      return -1;
   }
   return -1;
#else
   return -1;
#endif
}

/*
******************************************************************************
** FUNCTION:  changetolower
**
** PURPOSE:   Change the string to lower 
**
** PARAMETERS:
**    1. string 
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
static void changetolower(unsigned char *pp)
{
   unsigned char *p;
   p = pp;
   while(*p)
   {
      *p++ = (char)tolower((int)*p);
   }
}

/*
******************************************************************************
** FUNCTION:  secWriteNVRAMSkew
**
** PURPOSE:   Write time skew into NVRAM
**
** PARAMETERS:
**    1. time skew value
**    2. flag prov or cms (AS or AP)
**    3. realm to be hashed
**    4. length of realm
**
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
static void secWriteNVRAMSkew(
      int             skew_value,
      int             flag,
      char            *realm,
      unsigned long   realm_length)
{
#ifndef DEMO
   int           rc;
   int           ix;
   int           new_entry;
   MTA_KdcEntry  kdc_entry;
   MTA_CmsEntry  cms_entry;
   unsigned char lrealm[255];

   strcpy(lrealm, realm);
   changetolower(lrealm);

   switch (flag)
   {
      case AS_PROV_FLAG:
         cmNvSetProvKdcSkewTime(skew_value);
         break;
      case AP_PROV_FLAG:
         cmNvSetProvApSkewTime(skew_value);
         break;
      case AS_CMS_FLAG:
         /* find an available or existing index */
         ix = get_kdc_index(lrealm, realm_length);
         if ( ix < 0 )
         {
            LOG(("Write NVRAM Skew: no NVRAM kdcEntry left for %s.", realm));
            return;
         }
         kdc_entry.pRealmName = lrealm;
         kdc_entry.asSkewTime = skew_value;
         LOG(("Write NVRAM Skew: Writing KDC entry to NVRAM index %d.", ix));
         rc = cmNvSetKdcEntry(ix, &kdc_entry);
         if ( ! rc )
            LOG(("Write NVRAM Skew: could not write skew for %s index %d.", 
                     realm, ix));
         else
            LOG(("Write NVRAM Skew: stored skew for %s, index %d.", realm, ix));
         break;
      case AP_CMS_FLAG:
         /* get cms entry from NVRAM and overwrite the time skew value */
         /* realm here is actualy the cms fqdn */
         cms_entry.pKerberosTicket = tmp_tkt;
         cms_entry.pKerberosSessionKey = tmp_skey;
         memcpy(tmp_fqdn, lrealm, realm_length);
         cms_entry.pCmsFqdn = tmp_fqdn;
         /* find an available or existing index */
         ix = get_cms_index(lrealm, realm_length, &new_entry);
         if ( ix < 0 )
         {
            LOG(("Write NVRAM Skew: no NVRAM CmsEntry found for %s.", realm));
            return;
         }
         cmNvGetCmsEntry(ix, &cms_entry);
         cms_entry.apSkewTime = skew_value;
         cmNvSetCmsEntry(ix, &cms_entry);
         break;
      default:
         LOG(("Write NVRAM Skew: BAD FLAG %d.", flag));
         return;
         break;
   }
#else
   int           i;
   unsigned long tmp_len;

   /* find an open entry in the table */
   for (i=0;i<MAX_NVRAM_SKEW_TABLE_ENTRIES;i++)
      if ( skew_table[i].t_hash[0] == '\0')
         break;
   if ( i == MAX_NVRAM_SKEW_TABLE_ENTRIES )
      printf("*SEC PANIC* no more NVRAM Skew_table entries.\n");

   SecLibGetCryptoIface()->secCryptoCalcHash(HASH_MD5_TYPE, realm, realm_length,
         NULL, 0, (unsigned char *)(skew_table[i].t_hash), &tmp_len);

   skew_table[i].flag = (unsigned char)flag;
   skew_table[i].skew = skew_value;
#endif
}

/*
******************************************************************************
** FUNCTION:  secReadNVRAMSkew
**
** PURPOSE:   Read time skew into NVRAM
**
** PARAMETERS:
**    1. time skew value
**    2. flag prov or cms (AS or AP)
**    3. realm to be hashed
**    4. length of realm
**    5. if match, index to be used when writing
**
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
static int secReadNVRAMSkew(
      int             flag,
      char            *realm,
      unsigned long   realm_length)
{

#ifndef DEMO
   int           ix;
   int           rc;
   MTA_KdcEntry  kdc_entry;
   MTA_CmsEntry  cms_entry;
   unsigned char tmp_realm[255];
   unsigned char lrealm[255];

   strcpy(lrealm, realm);
   changetolower(lrealm);

   kdc_entry.pRealmName = tmp_realm;
   switch (flag)
   {
      case AS_PROV_FLAG:
         return (int)(cmNvGetProvKdcSkewTime());
         break;
      case AP_PROV_FLAG:
         return (int)(cmNvGetProvApSkewTime());
         break;
      case AS_CMS_FLAG:
         /* find index to this realm */
         ix = find_kdc_index(lrealm, realm_length);
         if ( ix < 0 )
         {
            LOG(("Read NVRAM Skew: no NVRAM kdcEntry found for %s.", realm));
            return 0;
         }
         LOG(("Retrieving AS CMS skew for realm %s, index %d ...", realm, ix));
         rc = cmNvGetKdcEntry(ix, &kdc_entry);
         if ( ! rc )
         {
            LOG(("Read NVRAM Skew: could not Read skew for %s.", realm));
            return 0;
         }
         else
         {
            LOG(("Read NVRAM Skew: Retrieved skew for %s, index %d.", 
                     kdc_entry.pRealmName, ix));
            return kdc_entry.asSkewTime;
         }
         break;
      case AP_CMS_FLAG:
         /* realm passed in is in fact cms fqdn */
         cms_entry.pKerberosTicket = tmp_tkt;
         cms_entry.pKerberosSessionKey = tmp_skey;
         cms_entry.pCmsFqdn = tmp_fqdn;
         ix = find_cms_index(lrealm, realm_length);
         if ( ix < 0 )
         {
            LOG(("Read NVRAM Skew: no NVRAM CmsEntry found for %s.", realm));
            return 0;
         }
         LOG(("Retrieving AP CMS skew for realm %s, index %d ...", realm, ix));
         rc = cmNvGetCmsEntry(ix, &cms_entry);
         if ( ! rc )
         {
            LOG(("Read NVRAM Skew: could not Read skew for %s.", realm));
            return 0;
         }
         else
         {
            LOG(("Read NVRAM Skew: Retrieved skew for %s, index %d.", 
                     cms_entry.pCmsFqdn, ix));
            return cms_entry.apSkewTime;
         }
         break;
      default:
         LOG(("Read NVRAM Skew BAD FLAG %d.", flag));
         return 0;
         break;
   }
#else
   unsigned char tmp_hash[MD5_SIZE];
   unsigned long tmp_len;
   int           i;


   SecLibGetCryptoIface()->secCryptoCalcHash(
         HASH_MD5_TYPE, realm, realm_length, NULL, 0,
         tmp_hash, &tmp_len);

   for (i=0;i<MAX_NVRAM_SKEW_TABLE_ENTRIES;i++)
   {
      if ( ! memcmp(skew_table[i].t_hash, tmp_hash, MD5_SIZE))
      {
         if ( skew_table[i].flag == flag )
         {
            return (skew_table[i].skew);
         }
      }
   }
   return 0;
#endif
}

/*
******************************************************************************
** FUNCTION:  secWriteNVRAMTkt
**
** PURPOSE:   Write ticket into NVRAM
**
** PARAMETERS:
**    1. asn.1 encoded ticket
**    2. flag prov or cms (AS or AP)
**    3. cms fqdn, ignored if prov
**    4. length of fqdn
**
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
static void secWriteNVRAMTkt(
      unsigned char   *tkt,
      int             tkt_len,
      int             flag,
      char            *fqdn,
      unsigned long   fqdn_length)
{
#ifndef DEMO
   MTA_CmsEntry      cms_entry;
   int               ix;
   int               new_entry;
   unsigned char     lfqdn[255];
   
   strcpy(lfqdn, fqdn);
   changetolower(lfqdn);

   switch (flag)
   {
      case AS_PROV_FLAG:
         cmNvSetSnmpV3KerberosTicket(tkt, tkt_len);
         break;
      case AS_CMS_FLAG:
         cms_entry.pKerberosTicket = tmp_tkt;
         cms_entry.pKerberosSessionKey = tmp_skey;
         cms_entry.pCmsFqdn = tmp_fqdn;
         /* find an available or existing index */
         ix = get_cms_index(lfqdn, fqdn_length, &new_entry);
         if ( ix < 0 )
         {
            LOG(("PANIC Write NVRAM tkt: no NVRAM CmsEntry left for %s.",fqdn));
            return;
         }
         if ( !new_entry )
         {
            cmNvGetCmsEntry(ix, &cms_entry);
         }
         cms_entry.pCmsFqdn = lfqdn;
         cms_entry.pCmsFqdn[fqdn_length] = '\0';
         memcpy(cms_entry.pKerberosTicket, tkt, tkt_len);
         cms_entry.kerberosTicketLength = tkt_len;
         if ( new_entry )
         {
            cms_entry.apSkewTime = 0;
            cms_entry.kerberosSessionKeyLength = 0;
         }
         cmNvSetCmsEntry(ix, &cms_entry);
         LOG(("CmsEntry written index=%d, fqdn=%s, len = %d.", 
                  ix, cms_entry.pCmsFqdn, cms_entry.kerberosSessionKeyLength));
         break;
      default:
         LOG(("secWriteNVRAMTkt BAD FLAG %d.", flag));
         break;
   }
#endif
}

/*
******************************************************************************
** FUNCTION:  secReadNVRAMTkt
**
** PURPOSE:   Read ticket into NVRAM
**
** PARAMETERS:
**    1. asn.1 encoded ticket
**    2. flag prov or cms (AS or AP)
**    3. cms fqdn, ignored if prov
**    4. length of fqdn
**
**
** RETURNS:
**    Length of the ticket
**
** NOTE:
**    Space for the ticket must be allocated before this call
**
******************************************************************************
*/
static unsigned long secReadNVRAMTkt(
      unsigned char   *tkt,
      unsigned long   tkt_len,
      int             flag,
      char            *fqdn,
      unsigned long   fqdn_length)
{
#ifndef DEMO
   MTA_CmsEntry      cms_entry;
   int               ix;
   BOOL              rc;
   unsigned char     lfqdn[255];

   strcpy(lfqdn, fqdn);
   changetolower(lfqdn);

   switch (flag)
   {
      case AS_PROV_FLAG:
         return (int)(cmNvGetSnmpV3KerberosTicket(tkt, tkt_len));
         break;
      case AS_CMS_FLAG:
         cms_entry.pKerberosTicket = tmp_tkt;
         cms_entry.pKerberosSessionKey = tmp_skey;
         cms_entry.pCmsFqdn = tmp_fqdn;
         ix = find_cms_index(lfqdn, fqdn_length);
         if ( ix < 0 )
         {
            LOG(("Read NVRAM Skew: no NVRAM CmsEntry found for %s.", fqdn));
            return 0;
         }
         LOG(("Retrieving tkt for fqdn %s, index %d ...", fqdn, ix));
         rc = cmNvGetCmsEntry(ix, &cms_entry);
         if ( ! rc )
         {
            LOG(("Read NVRAM tkt: could not Read tkt for %s.", fqdn));
            return 0;
         }
         else
         {
            LOG(("Read NVRAM tkt: Retrieved tkt for %s, index %d, len = %d.", 
                     cms_entry.pCmsFqdn, ix, cms_entry.kerberosTicketLength));
            memcpy(tkt, cms_entry.pKerberosTicket, cms_entry.kerberosTicketLength);
            return cms_entry.kerberosTicketLength;
         }
         break;
      default:
         LOG(("secReadNVRAMTkt BAD FLAG %d.", flag));
         break;
   }
#endif
   return 0;
}


/*
******************************************************************************
** FUNCTION:  secWriteNVRAMSessionKeyAndTktTimes
**
** PURPOSE:   Write session key and ticket times into NVRAM
**
** PARAMETERS:
**    1. session key always 24 bytes
**    2. length of session key
**    3. type of session key
**    4. flag prov or cms (AS or AP)
**    5. cms fqdn, ignored if prov
**    6. length of fqdn
**    7. ticket times (4 integers)
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
static void secWriteNVRAMSessionKeyAndTktTimes(
      unsigned char   *session_key,
      unsigned long   session_key_len,
      int             session_key_type,
      int             flag,
      char            *fqdn,
      unsigned long   fqdn_length,
      int             auth_time,
      int             start_time,
      int             end_time,
      int             renew_till)
{
#ifndef DEMO
   int              ix;
   int              new_entry;
   unsigned char    tmp_info[MAX_SESSION_KEY_LENGTH+16];
   MTA_CmsEntry     cms_entry;
   BOOL             rs;
   unsigned char    lfqdn[255];

   strcpy(lfqdn, fqdn);
   changetolower(lfqdn);

   switch (flag)
   {
      case AS_PROV_FLAG:
         /* cocatenate the session key and times into one buffere */
         memcpy(tmp_info, session_key, MAX_SESSION_KEY_LENGTH);
         memcpy(tmp_info+MAX_SESSION_KEY_LENGTH, &auth_time, 4); 
         memcpy(tmp_info+MAX_SESSION_KEY_LENGTH+4, &start_time, 4); 
         memcpy(tmp_info+MAX_SESSION_KEY_LENGTH+8, &end_time, 4); 
         memcpy(tmp_info+MAX_SESSION_KEY_LENGTH+12, &renew_till, 4); 
         cmNvSetSnmpV3SessionKeyAndTktTimes(tmp_info, 
               MAX_SESSION_KEY_LENGTH+16);
         break;
      case AS_CMS_FLAG:
         /* get cms entry from NVRAM and overwrite the time skew value */
         /* realm here is actualy the cms fqdn */
         cms_entry.pKerberosTicket = tmp_tkt;
         cms_entry.pKerberosSessionKey = tmp_skey;
         cms_entry.pCmsFqdn = tmp_fqdn;
         /* find an available or existing index */
         ix = get_cms_index(lfqdn, fqdn_length, &new_entry);
         if ( ix < 0 )
         {
            LOG(("Write NVRAM skey&times: no NVRAM CmsEntry left for %s.", 
                     fqdn));
            return;
         }
         cmNvGetCmsEntry(ix, &cms_entry);
         cms_entry.pCmsFqdn = lfqdn;
         cms_entry.pCmsFqdn[fqdn_length] = '\0';
         memcpy(cms_entry.pKerberosSessionKey, session_key, session_key_len);
         cms_entry.kerberosSessionKeyLength = session_key_len;
         cms_entry.ticketTime[0] = auth_time;
         cms_entry.ticketTime[1] = start_time;
         cms_entry.ticketTime[2] = end_time;
         cms_entry.ticketTime[3] = renew_till;
         rs = cmNvSetCmsEntry(ix, &cms_entry);
         if ( rs == FALSE )
         {
            LOG(("FAILED TO WRITE CMS ENTRY."));
            LOG(("CmsEntry index=%d, fqdn=%s, skey length =%d, tkt_len = %d.", 
                     ix, cms_entry.pCmsFqdn,
                     cms_entry.kerberosSessionKeyLength,
                     cms_entry.kerberosTicketLength));
         }
         else
         {
            LOG(("CmsEntry written index=%d, fqdn=%s, skey length =%d, tkt_len = %d.", 
                     ix, cms_entry.pCmsFqdn,
                     cms_entry.kerberosSessionKeyLength,
                     cms_entry.kerberosTicketLength));
         }
         break;
      default:
         LOG(("secWriteNVRAMSessionKeyAndTktTimes BAD FLAG %d.", flag));
         return;
         break;
   }
#endif
}

/*
******************************************************************************
** FUNCTION:  secReadNVRAMSessionKeyAndTktTimes
**
** PURPOSE:   Read session key and ticket times from NVRAM
**
** PARAMETERS:
**    1. session key
**    2. session key type
**    3. flag prov or cms (AS or AP)
**    4. cms fqdn, ignored if prov
**    5. length of fqdn
**    6. ticket times (4 integers)
**
**
** RETURNS:
**    length of the session key or 0
**
** NOTE:
**    caller must allocate space for session key
**
******************************************************************************
*/
static unsigned long secReadNVRAMSessionKeyAndTktTimes(
      unsigned char   *session_key,
      int             *session_key_type,
      int             flag,
      char            *fqdn,
      unsigned long   fqdn_length,
      int             *auth_time,
      int             *start_time,
      int             *end_time,
      int             *renew_till)
{
#ifndef DEMO
   int i;
   unsigned char     res_buf[MAX_SESSION_KEY_LENGTH+16];
   MTA_CmsEntry      cms_entry;
   int               ix;
   BOOL              rc;
   unsigned char     lfqdn[255];

   strcpy(lfqdn, fqdn);
   changetolower(lfqdn);

   switch (flag)
   {
      case AS_PROV_FLAG:
         rc = cmNvGetSnmpV3SessionKeyAndTktTimes(res_buf, 
               MAX_SESSION_KEY_LENGTH+16);
         
         if ( rc )
         {
            memcpy(session_key, res_buf, MAX_SESSION_KEY_LENGTH);
            memcpy(auth_time, res_buf+MAX_SESSION_KEY_LENGTH, 4);
            memcpy(start_time, res_buf+MAX_SESSION_KEY_LENGTH+4, 4);
            memcpy(end_time, res_buf+MAX_SESSION_KEY_LENGTH+8, 4);
            memcpy(renew_till, res_buf+MAX_SESSION_KEY_LENGTH+12, 4);

            return MAX_SESSION_KEY_LENGTH;
         }
         LOG(("Read NVRAM skey&time: no NVRAM SNMPv3 Session Key found for %s.", 
                  fqdn));
         return 0;
         break;
      case AS_CMS_FLAG:
         cms_entry.pKerberosTicket = tmp_tkt;
         cms_entry.pKerberosSessionKey = tmp_skey;
         cms_entry.pCmsFqdn = tmp_fqdn;
         ix = find_cms_index(lfqdn, fqdn_length);
         if ( ix < 0 )
         {
            LOG(("Read NVRAM skey&time: no NVRAM CmsEntry found for %s.", 
                     fqdn));
            return 0;
         }
         LOG(("Retrieving AP CMS skey&time for fqdn %s, index %d ...", 
                  fqdn, ix));
         rc = cmNvGetCmsEntry(ix, &cms_entry);
         if ( ! rc )
         {
            LOG(("Read NVRAM Skey&Times: could not Read skey for %s.", fqdn));
            return 0;
         }
         else
         {
            LOG(("Read NVRAM Skey&times: Retrieved skey for %s, index %d.", 
                     cms_entry.pCmsFqdn, ix));
            memcpy(session_key, cms_entry.pKerberosSessionKey, 
                  cms_entry.kerberosSessionKeyLength);
            *auth_time = cms_entry.ticketTime[0];
            *start_time = cms_entry.ticketTime[1];
            *end_time = cms_entry.ticketTime[2];
            *renew_till = cms_entry.ticketTime[3];
            LOG(("end_time = %d, skey_len = %d.", *end_time, 
                     cms_entry.kerberosSessionKeyLength));
            for ( i=0;i<cms_entry.kerberosSessionKeyLength;i++)
               printf("%x ", cms_entry.pKerberosSessionKey[i]);
            printf("\n");
            return cms_entry.kerberosSessionKeyLength;
         }
         break;
      default:
         LOG(("secWriteNVRAMSessionKeyAndTktTimes BAD FLAG %d.", flag));
         break;
   }
#endif

   return 0;
}


/*
******************************************************************************
** FUNCTION:  secReadNVRAMMTACertKeyData
**
** PURPOSE:   Read manufacture certificate from NVRAM
**
** PARAMETERS:
**    1. in: certificate type
**    2. out: certificate length
**    3. out: certificate
**
**
** RETURNS:
**    Success or failure
**
** NOTE:
**    caller responsible for allocating and freeing memory
******************************************************************************
*/
static unsigned long secReadNVRAMMTACertKeyData(
      SEC_CERT_TYPE   type,
      unsigned int    *len,
      unsigned char   *data)
{
#ifndef DEMO   
   BOOL            rc;
   unsigned int    cert_len = MAX_CERTIFICATE_LENGTH;
   unsigned int    key_len = MAX_CERT_KEY_LENGTH;

   
   switch (type)
   {
      case TEL_ROOT_CERT:
         rc = cmNvGetTelephonyRootCertificate(data, &cert_len);
         *len = cert_len;
         printf("TELROOT CERT len=%d, rc=%d.\n", *len, rc);
         return 1;
      case MTA_MAN_CERT:
         rc = cmNvGetMtaManufacturerCertificate(data, &cert_len);
         *len = cert_len;
         printf("MAN CERT len=%d, rc=%d.\n", *len, rc);
         return 1;
      case MTA_DEV_CERT:
         rc = cmNvGetMtaDeviceCertificate(data, &cert_len);
         *len = cert_len;
         printf("DEV CERT len=%d, rc=%d.\n", *len, rc);
         return 1;
      case MAT_DEV_MOD_KEY:
         rc = cmNvGetMtaDevPrivKeyModulus(data, &key_len);
         *len = key_len;
         printf("KEYMOD CERT len=%d, rc=%d.\n", *len, rc);
         return 1;
      case MTA_DEV_PRIV_KEY:
         rc = cmNvGetMtaDevPrivKeyExponent(data, &key_len);
         *len = key_len;
         printf("PRIVEXP CERT len=%d, rc=%d.\n", *len, rc);
         return 1;
      default:
         printf("NVRAM secReadNVRAMMTACertKeyData bad type %d.\n", type);
         return 0;
   }
#else
   switch (type)
   {
      case MTA_MAN_CERT:
         *len = MTA_man_cert.cert_len;
         memcpy(data, MTA_man_cert.cert, *len);
         return 1;
      case MTA_DEV_CERT:
         *len = MTA_dev_cert.cert_len;
         memcpy(data, MTA_dev_cert.cert, *len);
         return 1;
      case MAT_DEV_MOD_KEY:
         *len = MTA_device_modulus_key.key_len;
         memcpy(data, MTA_device_modulus_key.key, *len);
         return 1;
      case MTA_DEV_PRIV_KEY:
         *len = MTA_device_priv_key.key_len;
         memcpy(data, MTA_device_priv_key.key, *len);
         return 1;
      case TEL_ROOT_CERT:
         *len = TEL_root_cert.cert_len;
         memcpy(data, TEL_root_cert.cert, *len);
         return 1;
      default:
         printf("NVRAM secReadNVRAMMTACertKeyData bad type %d.\n", type);
         return 0;
   }
#endif
}

#ifdef DEMO
/*
******************************************************************************
** FUNCTION: secWriteNVRAMMTACertKeyData
**
** PURPOSE:   write a certificate or a key from NVRAM
**
** PARAMETERS:
**    1. in: certificate or key type
**    2. in: certificate or key length
**    3. in: certificate or key
**
**
** RETURNS:
**    Success or failure
**
** NOTE:
**    caller responsible for allocating and freeing memory
******************************************************************************
*/
static unsigned long secWriteNVRAMMTACertKeyData(
      SEC_CERT_TYPE   type,
      int             len,
      const unsigned char   *data)
{
   switch (type)
   {
      case MTA_MAN_CERT:
         if ( len > MAX_CERTIFICATE_LENGTH )
         {
            printf("NVRAM secWriteNVRAMMTACertKeyData Manufacture cert too big.\n", len);
            return 0;
         }
         MTA_man_cert.cert_len = len;
         memcpy(MTA_man_cert.cert, data, len);
         return 1;

      case MTA_DEV_CERT:
         if ( len > MAX_CERTIFICATE_LENGTH )
         {
            printf("NVRAM secWriteNVRAMMTACertKeyData Device cert too big.\n", len);
            return 0;
         }
         MTA_dev_cert.cert_len = len;
         memcpy(MTA_dev_cert.cert, data, len);
         return 1;

      case MAT_DEV_MOD_KEY:
         if ( len > MAX_KEY_LENGTH )
         {
            printf("NVRAM secWriteNVRAMMTACertKeyData Device modulus too big.\n", len);
            return 0;
         }
         MTA_device_modulus_key.key_len = len;
         memcpy(MTA_device_modulus_key.key, data, len);
         return 1;

      case MTA_DEV_PRIV_KEY:
         if ( len > MAX_KEY_LENGTH )
         {
            printf("NVRAM secWriteNVRAMMTACertKeyData Device private key too big.\n", len);
            return 0;
         }
         MTA_device_priv_key.key_len = len;
         memcpy(MTA_device_priv_key.key, data, len);
         return 1;

      case TEL_ROOT_CERT:
         if ( len > MAX_CERTIFICATE_LENGTH )
         {
            printf("NVRAM secWriteNVRAMMTACertKeyData Telephony Root Cert too big.\n", len);
            return 0;
         }
         TEL_root_cert.cert_len = len;
         memcpy(TEL_root_cert.cert, data, len);
         return 1;

      default:
         printf("NVRAM secWriteNVRAMMTACertKeyData bad type %d.\n", type);
         return 0;
   }
#else
   return 1;
#endif
}
