/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: net.h
*    Creation Date: 18 Aug 2000 (v0.00)
*    VSS Info:
*        $Revision: 1.2 $
*        $Date: 2005/02/17 21:13:21 $
*
****************************************************************************
*    Description:
*
*     This header file contains prototypes for NET utililty functions
*
****************************************************************************/

#ifndef _BCMNET_H
#define _BCMNET_H


#if defined(__cplusplus)
extern "C"
{
#endif

#include <hausenv.h>
#include <ChKerbNamespace.h>
/*
*****************************************************************************
** TYPEDEFS AND CONSTANTS
*****************************************************************************
*/

/* NET module API return code */
typedef enum
{
   NET_STATUS_SOCKET_ERROR = -1,    /* Socket error */
   NET_STATUS_OK,                   /* Success */
   NET_STATUS_ERROR                 /* General error code */
} NET_STATUS;

/*
*****************************************************************************
** FUNCTION:   bcmNetInit
**
** PURPOSE:    Initialize the NET module
**
** PARAMETERS: none
**
**
** RETURNS:    NET_STATUS_OK or error code
**
** NOTE:
*****************************************************************************
*/
NET_STATUS bcmNetInit( void );


/*
*****************************************************************************
** FUNCTION:   bcmNetDeinit
**
** PURPOSE:    Shutdown the NET module
**
** PARAMETERS: none
**
** RETURNS:    NET_STATUS_OK or error code
**             This function should only return an error under catastrophic
**             circumstances. i.e. Something that cannot be fixed by re-invoking
**             the module initialization function.
**
** NOTE:
*****************************************************************************
*/
NET_STATUS bcmNetDeinit( void );


/*
*****************************************************************************
** FUNCTION:   bcmNetNumToIpAddr()
**
** PURPOSE:    Converts IPADDR type (four byte int) to a dotted decimal
**             notation IP address string. e.g. converts 0x0a010005 to
**             "10.1.0.5".
**
** PARAMETERS: ipaddr - IPADDR IP address
**             buf    - string containing IP Address (output)
**
** RETURNS:    NET_STATUS_OK or error code
**
** NOTE:
*****************************************************************************
*/
NET_STATUS bcmNetNumToIpAddr( IPADDR ipaddr, char *buf );


/*
*****************************************************************************
** FUNCTION:   bcmNetIpAddrToNum()
**
** PURPOSE:    Converts character represented IP address (dotted decimal notation)
**             or fully qualified domain name (FQDN) to an IPADDR type (four byte
**             integer).
**
** PARAMETERS: string - character string with IP address/FQDN
**             ipAddr - ip address (four byte integer)
**
** RETURNS:    NET_STATUS_OK or error
**
** NOTE:       This function will first try to convert the passed in string
**             assuming that it is a dotted decimal notion IP address. If
**             this conversion fails, it will perform a DNS lookup. If this
**             fails, it will then return an error.
**
**             Use of this function is deprecated, and is only provided for
**             backwards compatibility. bcmNetGetHostAddrs should be used instead.
**
*****************************************************************************
*/
NET_STATUS  bcmNetIpAddrToNum( const char *string, IPADDR *ipAddr );

/*
****************************************************************************
** FUNCTION:   bcmNetGetHostAddrs
**
** PURPOSE:    Given the hostname, this function finds corresponding IP
**             addresses and returns them.
**
** PARAMETERS: hostName [IN]        - host name to resolve into IP addresses
**             ipAddr   [OUT]       - array of resolved IP addresses (in
**                                     host byte order). Memory must be
**                                     allocated by caller.
**             numIpAddrs [IN/OUT]  - On input, points to the array length
**                                     of 'ipAddr'. On output, returns the
**                                     number of resolved IP addresses contained
**                                     in the 'ipAddr' array.
**             useCache [IN]        - indicates whether to use the DNS cache
**
** RETURNS:    NET_STATUS_OK or error code
**
** NOTE:       The 'hostName' may be either a fully qualified domain name or
**             a dotted decimal notation IP address string.
**
*****************************************************************************
*/
NET_STATUS bcmNetGetHostAddrs( char *hostName, IPADDR *ipAddr, int *numIpAddrs, BOOL useCache );

/*
*****************************************************************************
** FUNCTION:   bcmNetUdpOpen
**
** PURPOSE:    Open UDP socket
**
** PARAMETERS: localPort  - local port of socket
**             bBlockMode - blocking or non-blocking mode of the socket
**             sd         - socket descriptor (output)
**
** RETURNS:    NET_STATUS_OK or error code
**
*****************************************************************************
*/
NET_STATUS bcmNetUdpOpen( UINT16 localPort, int bBlockMode, int *sd );


/*
*****************************************************************************
** FUNCTION:   bcmNetUdpRecv
**
** PURPOSE:    Receive UDP socket
**
** PARAMETERS: udpSock  - udp socket port
**             buf      - pointer to receive buffer
**             length   - length of receive buffer
**             rAddr    - pointer to remote udp socket address (in host-byte order)
**             rPort    - pointer to remote udp socket port (in host-byte order)
**             nRxBytes - number of received bytes
**
** RETURNS:    NET_STATUS_OK or error code
**
*****************************************************************************
*/
NET_STATUS bcmNetUdpRecv( int             udpSock,
                          char            *buf,
                          int             length,
                          unsigned long   *raddr,
                          UINT16          *rport,
                          int             *nRxBytes );

/*
*****************************************************************************
** FUNCTION:   bcmNetUdpSend
**
** PURPOSE:    Send UDP socket
**
** PARAMETERS: udpSock  - udp socket port
**             buf      - pointer to transmit buffer
**             length   - length of transmit buffer
**             remAddr  - remote udp socket address (in host-byte order)
**             remPort  - remote udp socket port (in host-byte order)
**             nTxBytes - number of bytes transmitted (output)
**
** RETURNS:    NET_STATUS_OK or error code
**
*****************************************************************************
*/
NET_STATUS bcmNetUdpSend( int             udpSock,
                          char            *buf,
                          int             length,
                          unsigned long   remAddr,
                          UINT16          remPort,
                          int             *nTxBytes );

/*
*****************************************************************************
** FUNCTION:   bcmNetUdpClose
**
** PURPOSE:    Close UDP socket
**
** PARAMETERS: sockDesc - udp socket description
**
** RETURNS:    NET_STATUS_OK or error code
**
*****************************************************************************
*/
NET_STATUS bcmNetUdpClose( int sockDesc );

/*
*****************************************************************************
** FUNCTION:   bcmNetGetOwnIpAddress()
**
** PURPOSE:    Retrieves the IP address of a given IP stack number
**
** PARAMETERS: netIface - IP Interface Number
**             ipAddr - IP address of interface number
**
** RETURNS:    NET_STATUS_OK or error code
**
** NOTE:       Parameter always defaults to Iface# = 1 if set to 0.
*****************************************************************************
*/
NET_STATUS bcmNetGetOwnIpAddress( int netIface, IPADDR *ipAddr );

/*
*****************************************************************************
** FUNCTION:   bcmNetGetPortNum
**
** PURPOSE:    Extracts the Port Number part from the IP-Address string.
**             IP-Address string should be like <IP-str>:<port_num>
**
** PARAMETERS: ipAddrStr [IN]  - IP Addr string (or FQDN)
**             port      [OUT] - port number
**
** RETURNS:    NET_STATUS_OK or error code
**
** NOTE:
*****************************************************************************
*/
NET_STATUS bcmNetGetPortNum( const char* ipAddrStr, UINT32 *port );


#if defined(__cplusplus)
}
#endif

#endif /* _BCMNET_H */

